from typing import Generic, TypeVar

_InstanceType = TypeVar('_InstanceType')
_ReturnType = TypeVar('_ReturnType')

_PipelineStepType1 = TypeVar('_PipelineStepType1')
_PipelineStepType2 = TypeVar('_PipelineStepType2')
_PipelineStepType3 = TypeVar('_PipelineStepType3')
_PipelineStepType4 = TypeVar('_PipelineStepType4')
_PipelineStepType5 = TypeVar('_PipelineStepType5')
_PipelineStepType6 = TypeVar('_PipelineStepType6')
_PipelineStepType7 = TypeVar('_PipelineStepType7')
_PipelineStepType8 = TypeVar('_PipelineStepType8')
_PipelineStepType9 = TypeVar('_PipelineStepType9')
_PipelineStepType10 = TypeVar('_PipelineStepType10')
_PipelineStepType11 = TypeVar('_PipelineStepType11')
_PipelineStepType12 = TypeVar('_PipelineStepType12')
_PipelineStepType13 = TypeVar('_PipelineStepType13')
_PipelineStepType14 = TypeVar('_PipelineStepType14')
_PipelineStepType15 = TypeVar('_PipelineStepType15')
_PipelineStepType16 = TypeVar('_PipelineStepType16')
_PipelineStepType17 = TypeVar('_PipelineStepType17')
_PipelineStepType18 = TypeVar('_PipelineStepType18')
_PipelineStepType19 = TypeVar('_PipelineStepType19')
_PipelineStepType20 = TypeVar('_PipelineStepType20')

class _Pipe(
    Generic[
        _InstanceType,
        _ReturnType,
        _PipelineStepType1,
        _PipelineStepType2,
        _PipelineStepType3,
        _PipelineStepType4,
        _PipelineStepType5,
        _PipelineStepType6,
        _PipelineStepType7,
        _PipelineStepType8,
        _PipelineStepType9,
        _PipelineStepType10,
        _PipelineStepType11,
        _PipelineStepType12,
        _PipelineStepType13,
        _PipelineStepType14,
        _PipelineStepType15,
        _PipelineStepType16,
        _PipelineStepType17,
        _PipelineStepType18,
        _PipelineStepType19,
        _PipelineStepType20,
    ]
):
    def __init__(  # noqa: WPS451
        self,
        functions: tuple[
            _PipelineStepType1,
            _PipelineStepType2,
            _PipelineStepType3,
            _PipelineStepType4,
            _PipelineStepType5,
            _PipelineStepType6,
            _PipelineStepType7,
            _PipelineStepType8,
            _PipelineStepType9,
            _PipelineStepType10,
            _PipelineStepType11,
            _PipelineStepType12,
            _PipelineStepType13,
            _PipelineStepType14,
            _PipelineStepType15,
            _PipelineStepType16,
            _PipelineStepType17,
            _PipelineStepType18,
            _PipelineStepType19,
            _PipelineStepType20,
        ],
        /,
    ) -> None: ...
    def __call__(self, instance: _InstanceType) -> _ReturnType: ...

def pipe(  # noqa: WPS451
    function1: _PipelineStepType1,
    function2: _PipelineStepType2 = ...,
    function3: _PipelineStepType3 = ...,
    function4: _PipelineStepType4 = ...,
    function5: _PipelineStepType5 = ...,
    function6: _PipelineStepType6 = ...,
    function7: _PipelineStepType7 = ...,
    function8: _PipelineStepType8 = ...,
    function9: _PipelineStepType9 = ...,
    function10: _PipelineStepType10 = ...,
    function11: _PipelineStepType11 = ...,
    function12: _PipelineStepType12 = ...,
    function13: _PipelineStepType13 = ...,
    function14: _PipelineStepType14 = ...,
    function15: _PipelineStepType15 = ...,
    function16: _PipelineStepType16 = ...,
    function17: _PipelineStepType17 = ...,
    function18: _PipelineStepType18 = ...,
    function19: _PipelineStepType19 = ...,
    function20: _PipelineStepType20 = ...,
    /,
) -> _Pipe[
    _InstanceType,
    _ReturnType,
    _PipelineStepType1,
    _PipelineStepType2,
    _PipelineStepType3,
    _PipelineStepType4,
    _PipelineStepType5,
    _PipelineStepType6,
    _PipelineStepType7,
    _PipelineStepType8,
    _PipelineStepType9,
    _PipelineStepType10,
    _PipelineStepType11,
    _PipelineStepType12,
    _PipelineStepType13,
    _PipelineStepType14,
    _PipelineStepType15,
    _PipelineStepType16,
    _PipelineStepType17,
    _PipelineStepType18,
    _PipelineStepType19,
    _PipelineStepType20,
]: ...



import pytest

from returns.context import NoDeps, ReaderFutureResult
from returns.io import IOFailure, IOSuccess
from returns.pipeline import managed
from returns.result import Failure, Result, Success


def _acquire_success() -> ReaderFutureResult[str, str, NoDeps]:
    return ReaderFutureResult.from_value('acquire success')


def _acquire_failure() -> ReaderFutureResult[str, str, NoDeps]:
    return ReaderFutureResult.from_failure('acquire failure')


def _use_success(inner_value: str) -> ReaderFutureResult[str, str, NoDeps]:
    return ReaderFutureResult.from_value('use success')


def _use_failure(inner_value: str) -> ReaderFutureResult[str, str, NoDeps]:
    return ReaderFutureResult.from_failure('use failure')


class _ReleaseSuccess:
    def __init__(self, logs: list[tuple[str, Result[str, str]]]) -> None:
        self._logs = logs

    def __call__(
        self,
        inner_value: str,
        use_result: Result[str, str],
    ) -> ReaderFutureResult[None, str, NoDeps]:
        self._logs.append((inner_value, use_result))
        return ReaderFutureResult.from_value(None)


class _ReleaseFailure:
    def __init__(self, logs: list[tuple[str, Result[str, str]]]) -> None:
        self._logs = logs

    def __call__(
        self,
        inner_value: str,
        use_result: Result[str, str],
    ) -> ReaderFutureResult[None, str, NoDeps]:
        return ReaderFutureResult.from_failure('release failure')


@pytest.mark.anyio
@pytest.mark.parametrize(
    ('acquire', 'use', 'release', 'final_result', 'log'),
    [
        # Acquire success:
        (
            _acquire_success,
            _use_success,
            _ReleaseSuccess,
            IOSuccess('use success'),
            [('acquire success', Success('use success'))],
        ),
        (
            _acquire_success,
            _use_success,
            _ReleaseFailure,
            IOFailure('release failure'),
            [],
        ),
        (
            _acquire_success,
            _use_failure,
            _ReleaseSuccess,
            IOFailure('use failure'),
            [('acquire success', Failure('use failure'))],
        ),
        (
            _acquire_success,
            _use_failure,
            _ReleaseFailure,
            IOFailure('release failure'),
            [],
        ),
        # Acquire failure:
        (
            _acquire_failure,
            _use_success,
            _ReleaseSuccess,
            IOFailure('acquire failure'),
            [],
        ),
        (
            _acquire_failure,
            _use_failure,
            _ReleaseSuccess,
            IOFailure('acquire failure'),
            [],
        ),
        (
            _acquire_failure,
            _use_success,
            _ReleaseFailure,
            IOFailure('acquire failure'),
            [],
        ),
        (
            _acquire_failure,
            _use_failure,
            _ReleaseFailure,
            IOFailure('acquire failure'),
            [],
        ),
    ],
)
async def test_all_success(acquire, use, release, final_result, log):
    """Ensures that managed works as intended."""
    pipeline_logs: list[tuple[str, Result[str, str]]] = []
    pipeline_result = managed(
        use,
        release(pipeline_logs),
    )(acquire())

    assert await pipeline_result(ReaderFutureResult.no_args) == final_result
    assert pipeline_logs == log


@pytest.mark.anyio
async def test_full_typing():
    """This test is here to be a case for typing."""
    logs: list[tuple[str, Result[str, str]]] = []
    pipeline_result = managed(
        _use_success,
        _ReleaseSuccess(logs),
    )(_acquire_success())
    inner = pipeline_result(ReaderFutureResult.no_args)

    assert await inner == IOSuccess('use success')
    assert logs == [('acquire success', Success('use success'))]



from returns.methods.cond import cond as cond
from returns.methods.partition import partition as partition
from returns.methods.unwrap_or_failure import (
    unwrap_or_failure as unwrap_or_failure,
)



import pytest

from returns.primitives.exceptions import UnwrapFailedError
from returns.result import Failure, Success


def test_unwrap_success():
    """Ensures that unwrap works for Success container."""
    with pytest.raises(UnwrapFailedError):
        assert Success(5).failure()


def test_unwrap_failure():
    """Ensures that unwrap works for Failure container."""
    assert Failure(5).failure() == 5



from copy import copy, deepcopy

import pytest

from returns.context import RequiresContext
from returns.primitives.exceptions import ImmutableStateError


def test_requires_context_immutable() -> None:
    """Ensures that Context is immutable."""
    with pytest.raises(ImmutableStateError):
        RequiresContext.from_value(1).abc = 1


def test_requires_context_immutable_copy() -> None:
    """Ensures that Context returns it self when passed to copy function."""
    context = RequiresContext.from_value(1)
    assert context is copy(context)


def test_requires_context_immutable_deepcopy() -> None:
    """Ensures that Context returns it self when passed to deepcopy function."""
    context = RequiresContext.from_value(1)
    assert context is deepcopy(context)



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, TypeVar

from returns.interfaces.specific.reader_result import ReaderResultLikeN
from returns.primitives.hkt import Kinded, KindN, kinded

if TYPE_CHECKING:
    from returns.context import ReaderResult  # noqa: WPS433

_FirstType = TypeVar('_FirstType')
_SecondType = TypeVar('_SecondType')
_ThirdType = TypeVar('_ThirdType')
_UpdatedType = TypeVar('_UpdatedType')

_ReaderResultLikeKind = TypeVar(
    '_ReaderResultLikeKind',
    bound=ReaderResultLikeN,
)


def bind_context_result(
    function: Callable[
        [_FirstType],
        ReaderResult[_UpdatedType, _SecondType, _ThirdType],
    ],
) -> Kinded[
    Callable[
        [KindN[_ReaderResultLikeKind, _FirstType, _SecondType, _ThirdType]],
        KindN[_ReaderResultLikeKind, _UpdatedType, _SecondType, _ThirdType],
    ]
]:
    """
    Composes successful container with a function that returns a container.

    In other words, it modifies the function's
    signature from:
    ``a -> ReaderResult[b, c, e]``
    to:
    ``Container[a, c, e] -> Container[b, c, e]``

    .. code:: python

      >>> from returns.pointfree import bind_context_result
      >>> from returns.context import ReaderIOResult, ReaderResult
      >>> from returns.io import IOSuccess, IOFailure

      >>> def example(argument: int) -> ReaderResult[int, str, str]:
      ...     return ReaderResult.from_value(argument + 1)

      >>> assert bind_context_result(example)(
      ...     ReaderIOResult.from_value(1),
      ... )(...) == IOSuccess(2)
      >>> assert bind_context_result(example)(
      ...     ReaderIOResult.from_failure('a'),
      ... )(...) == IOFailure('a')

    """

    @kinded
    def factory(
        container: KindN[
            _ReaderResultLikeKind,
            _FirstType,
            _SecondType,
            _ThirdType,
        ],
    ) -> KindN[_ReaderResultLikeKind, _UpdatedType, _SecondType, _ThirdType]:
        return container.bind_context_result(function)

    return factory



from copy import copy, deepcopy

import pytest

from returns.context import RequiresContextIOResult
from returns.primitives.exceptions import ImmutableStateError


def test_requires_context_result_immutable():
    """Ensures that container is immutable."""
    with pytest.raises(ImmutableStateError):
        RequiresContextIOResult.from_value(1).abc = 1


def test_requires_context_result_immutable_copy():
    """Ensures that helper returns it self when passed to copy function."""
    context_ioresult = RequiresContextIOResult.from_value(1)
    assert context_ioresult is copy(context_ioresult)


def test_requires_context_result_immutable_deepcopy():  # noqa: WPS118
    """Ensures that helper returns it self when passed to deepcopy function."""
    requires_context = RequiresContextIOResult.from_value(1)
    assert requires_context is deepcopy(requires_context)



from returns._internal.pipeline.flow import flow


def pipe(*functions):
    """
    Allows to compose a value and up to 7 functions that use this value.

    We use a custom ``mypy`` plugin to make sure types are correct.
    Otherwise, it is currently impossible to properly type this function.

    Each next function uses the previous result as an input parameter.
    Here's how it should be used:

    .. code:: python

       >>> from returns.pipeline import pipe

       >>> # => executes: str(float(int('1')))
       >>> assert pipe(int, float, str)('1') == '1.0'

    This function is closely related
    to :func:`pipe <returns._internal.pipeline.flow.flow>`:

    .. code:: python

      >>> from returns.pipeline import flow
      >>> assert pipe(int, float, str)('1') == flow('1', int, float, str)

    See also:
        - https://stackoverflow.com/a/41585450/4842742
        - https://github.com/gcanti/fp-ts/blob/master/src/pipeable.ts

    """
    return lambda instance: flow(instance, *functions)



from abc import abstractmethod
from collections.abc import Callable
from typing import Never, TypeVar, final

from returns.interfaces import bindable, equable, lashable, swappable
from returns.primitives.container import BaseContainer, container_equality
from returns.primitives.hkt import Kind2, KindN, SupportsKind2, dekind

_FirstType = TypeVar('_FirstType')
_SecondType = TypeVar('_SecondType')
_ThirdType = TypeVar('_ThirdType')

_NewFirstType = TypeVar('_NewFirstType')
_NewSecondType = TypeVar('_NewSecondType')

_PairLikeKind = TypeVar('_PairLikeKind', bound='PairLikeN')


class PairLikeN(
    bindable.BindableN[_FirstType, _SecondType, _ThirdType],
    swappable.SwappableN[_FirstType, _SecondType, _ThirdType],
    lashable.LashableN[_FirstType, _SecondType, _ThirdType],
    equable.Equable,
):
    """Special interface for types that look like a ``Pair``."""

    @abstractmethod
    def pair(
        self: _PairLikeKind,
        function: Callable[
            [_FirstType, _SecondType],
            KindN[_PairLikeKind, _NewFirstType, _NewSecondType, _ThirdType],
        ],
    ) -> KindN[_PairLikeKind, _NewFirstType, _NewSecondType, _ThirdType]:
        """Allows to work with both arguments at the same time."""

    @classmethod
    @abstractmethod
    def from_paired(
        cls: type[_PairLikeKind],
        first: _NewFirstType,
        second: _NewSecondType,
    ) -> KindN[_PairLikeKind, _NewFirstType, _NewSecondType, _ThirdType]:
        """Allows to create a PairLikeN from just two values."""

    @classmethod
    @abstractmethod
    def from_unpaired(
        cls: type[_PairLikeKind],
        inner_value: _NewFirstType,
    ) -> KindN[_PairLikeKind, _NewFirstType, _NewFirstType, _ThirdType]:
        """Allows to create a PairLikeN from just a single object."""


PairLike2 = PairLikeN[_FirstType, _SecondType, Never]
PairLike3 = PairLikeN[_FirstType, _SecondType, _ThirdType]


@final
class Pair(
    BaseContainer,
    SupportsKind2['Pair', _FirstType, _SecondType],
    PairLike2[_FirstType, _SecondType],
):
    """
    A type that represents a pair of something.

    Like to coordinates ``(x, y)`` or two best friends.
    Or a question and an answer.

    """

    def __init__(
        self,
        inner_value: tuple[_FirstType, _SecondType],
    ) -> None:
        """Saves passed tuple as ``._inner_value`` inside this instance."""
        super().__init__(inner_value)

    # `Equable` part:

    equals = container_equality  # we already have this defined for all types

    # `Mappable` part via `BiMappable`:

    def map(
        self,
        function: Callable[[_FirstType], _NewFirstType],
    ) -> 'Pair[_NewFirstType, _SecondType]':
        """
        Changes the first type with a pure function.

        >>> assert Pair((1, 2)).map(str) == Pair(('1', 2))

        """
        return Pair((function(self._inner_value[0]), self._inner_value[1]))

    # `BindableN` part:

    def bind(
        self,
        function: Callable[
            [_FirstType],
            Kind2['Pair', _NewFirstType, _SecondType],
        ],
    ) -> 'Pair[_NewFirstType, _SecondType]':
        """
        Changes the first type with a function returning another ``Pair``.

        >>> def bindable(first: int) -> Pair[str, str]:
        ...     return Pair((str(first), ''))

        >>> assert Pair((1, 'b')).bind(bindable) == Pair(('1', ''))

        """
        return dekind(function(self._inner_value[0]))

    # `AltableN` part via `BiMappableN`:

    def alt(
        self,
        function: Callable[[_SecondType], _NewSecondType],
    ) -> 'Pair[_FirstType, _NewSecondType]':
        """
        Changes the second type with a pure function.

        >>> assert Pair((1, 2)).alt(str) == Pair((1, '2'))

        """
        return Pair((self._inner_value[0], function(self._inner_value[1])))

    # `LashableN` part:

    def lash(
        self,
        function: Callable[
            [_SecondType],
            Kind2['Pair', _FirstType, _NewSecondType],
        ],
    ) -> 'Pair[_FirstType, _NewSecondType]':
        """
        Changes the second type with a function returning ``Pair``.

        >>> def lashable(second: int) -> Pair[str, str]:
        ...     return Pair(('', str(second)))

        >>> assert Pair(('a', 2)).lash(lashable) == Pair(('', '2'))

        """
        return dekind(function(self._inner_value[1]))

    # `SwappableN` part:

    def swap(self) -> 'Pair[_SecondType, _FirstType]':
        """
        Swaps ``Pair`` elements.

        >>> assert Pair((1, 2)).swap() == Pair((2, 1))

        """
        return Pair((self._inner_value[1], self._inner_value[0]))

    # `PairLikeN` part:

    def pair(
        self,
        function: Callable[
            [_FirstType, _SecondType],
            Kind2['Pair', _NewFirstType, _NewSecondType],
        ],
    ) -> 'Pair[_NewFirstType, _NewSecondType]':
        """
        Creates a new ``Pair`` from an existing one via a passed function.

        >>> def min_max(first: int, second: int) -> Pair[int, int]:
        ...     return Pair((min(first, second), max(first, second)))

        >>> assert Pair((2, 1)).pair(min_max) == Pair((1, 2))
        >>> assert Pair((1, 2)).pair(min_max) == Pair((1, 2))

        """
        return dekind(function(self._inner_value[0], self._inner_value[1]))

    @classmethod
    def from_paired(
        cls,
        first: _NewFirstType,
        second: _NewSecondType,
    ) -> 'Pair[_NewFirstType, _NewSecondType]':
        """
        Creates a new pair from two values.

        >>> assert Pair.from_paired(1, 2) == Pair((1, 2))

        """
        return Pair((first, second))

    @classmethod
    def from_unpaired(
        cls,
        inner_value: _NewFirstType,
    ) -> 'Pair[_NewFirstType, _NewFirstType]':
        """
        Creates a new pair from a single value.

        >>> assert Pair.from_unpaired(1) == Pair((1, 1))

        """
        return Pair((inner_value, inner_value))



from typing import Final

# Constant fullnames for typechecking
# ===================================

#: Used for typed ``partial`` function.
TYPED_PARTIAL_FUNCTION: Final = 'returns.curry.partial'

#: Used for typed ``curry`` decorator.
TYPED_CURRY_FUNCTION: Final = 'returns.curry.curry'

#: Used for typed ``flow`` call.
TYPED_FLOW_FUNCTION: Final = 'returns._internal.pipeline.flow.flow'

#: Used for typed ``pipe`` call.
TYPED_PIPE_FUNCTION: Final = 'returns._internal.pipeline.pipe.pipe'
TYPED_PIPE_METHOD: Final = 'returns._internal.pipeline.pipe._Pipe.__call__'

#: Used for HKT emulation.
TYPED_KINDN: Final = 'returns.primitives.hkt.KindN'
TYPED_KINDN_ACCESS: Final = f'{TYPED_KINDN}.'
TYPED_KIND_DEKIND: Final = 'returns.primitives.hkt.dekind'
TYPED_KIND_KINDED_CALL: Final = 'returns.primitives.hkt.Kinded.__call__'
TYPED_KIND_KINDED_GET: Final = 'returns.primitives.hkt.Kinded.__get__'

#: Used for :ref:`do-notation`.
DO_NOTATION_METHODS: Final = (
    # Just validation:
    'returns.io.IO.do',
    'returns.maybe.Maybe.do',
    'returns.future.Future.do',
    # Also infer error types:
    'returns.result.Result.do',
    'returns.io.IOResult.do',
    'returns.future.FutureResult.do',
)



import pytest

from returns.context import (
    RequiresContext,
    RequiresContextFutureResult,
    RequiresContextIOResult,
    RequiresContextResult,
)
from returns.converters import flatten
from returns.future import Future, FutureResult
from returns.io import IO, IOFailure, IOSuccess
from returns.maybe import Nothing, Some
from returns.result import Failure, Success


@pytest.mark.parametrize(
    ('container', 'merged'),
    [
        # Flattens:
        (IO(IO(1)), IO(1)),
        (Success(Success({})), Success({})),
        (IOSuccess(IOSuccess(1)), IOSuccess(1)),
        (Some(Some(None)), Some(None)),
        (Some(Some([])), Some([])),
        # Nope:
        (Nothing, Nothing),
        (Failure(Failure('a')), Failure(Failure('a'))),
        (Failure(Success('a')), Failure(Success('a'))),
        (IOFailure(IOFailure('a')), IOFailure(IOFailure('a'))),
        (IOFailure(IOSuccess('a')), IOFailure(IOSuccess('a'))),
    ],
)
def test_flatten(container, merged):
    """Ensures that `flatten` is always returning the correct type."""
    assert flatten(container) == merged


@pytest.mark.parametrize(
    ('container', 'merged'),
    [
        (
            RequiresContextResult.from_value(
                RequiresContextResult.from_value(1),
            ),
            RequiresContextResult.from_value(1),
        ),
        (
            RequiresContextIOResult.from_value(
                RequiresContextIOResult.from_value(1),
            ),
            RequiresContextIOResult.from_value(1),
        ),
        (
            RequiresContext.from_value(RequiresContext.from_value(1)),
            RequiresContext.from_value(1),
        ),
    ],
)
def test_flatten_context(container, merged):
    """Ensures that `flatten` is always returning the correct type."""
    assert flatten(container)(...) == merged(...)


@pytest.mark.anyio
async def test_flatten_future(subtests):
    """Ensures that `flatten` is always returning the correct type."""
    futures = [
        # Flattens:
        (Future.from_value(Future.from_value(1)), Future.from_value(1)),
        (
            FutureResult.from_value(FutureResult.from_value(1)),
            FutureResult.from_value(1),
        ),
    ]

    for container, merged in futures:
        with subtests.test(container=container, merged=merged):
            assert await flatten(container) == await merged  # type: ignore


@pytest.mark.anyio
async def test_flatten_context_future_result(subtests):
    """Ensures that `flatten` is always returning the correct type."""
    futures = [
        # Flattens:
        (
            RequiresContextFutureResult.from_value(
                RequiresContextFutureResult.from_value(1),
            ),
            RequiresContextFutureResult.from_value(1),
        ),
    ]

    for container, merged in futures:
        with subtests.test(container=container, merged=merged):
            assert await flatten(
                container,
            )(...) == await merged(...)


@pytest.mark.anyio
async def test_non_flatten_future(subtests):
    """Ensures that `flatten` is always returning the correct type."""
    futures = [
        # Not flattens:
        FutureResult.from_failure(FutureResult.from_failure(1)),
        FutureResult.from_failure(FutureResult.from_value(1)),
    ]

    for cont in futures:
        with subtests.test(container=cont):
            assert isinstance(
                (await flatten(cont)).failure()._inner_value,  # noqa: SLF001
                cont.__class__,
            )


@pytest.mark.anyio
async def test_non_flatten_context_future_result(subtests):
    """Ensures that `flatten` is always returning the correct type."""
    futures = [
        # Not flattens:
        RequiresContextFutureResult.from_failure(
            RequiresContextFutureResult.from_failure(1),
        ),
        RequiresContextFutureResult.from_failure(
            RequiresContextFutureResult.from_value(1),
        ),
    ]

    for cont in futures:
        with subtests.test(container=cont):
            inner = await flatten(cont)(...)
            assert isinstance(
                inner.failure()._inner_value,  # noqa: SLF001
                cont.__class__,
            )



from returns.io import IO


def test_equals():
    """Ensures that ``.equals`` method works correctly."""
    assert IO(1).equals(IO(1))
    assert IO(1).equals(IO.from_value(1))


def test_not_equals():
    """Ensures that ``.equals`` method works correctly."""
    assert not IO(1).equals(IO('a'))


def test_equality():
    """Ensures that containers can be compared."""
    assert IO(1) == IO(1)
    assert str(IO(2)) == '<IO: 2>'
    assert hash(IO((1, 2, 3)))


def test_nonequality():
    """Ensures that containers are not compared to regular values."""
    assert IO(1) != 1
    assert IO(2) is not IO(2)
    assert IO('a') != IO('b')



from hypothesis import strategies as st
from test_hypothesis.test_laws import test_custom_type_applicative

from returns.contrib.hypothesis.laws import check_all_laws

container_type = test_custom_type_applicative._Wrapper  # noqa: SLF001

check_all_laws(
    container_type,
    container_strategy=st.builds(container_type, st.integers()),
)



from collections.abc import Callable
from typing import TypeVar

from returns.interfaces.mappable import MappableN
from returns.primitives.hkt import Kinded, KindN, kinded

_FirstType = TypeVar('_FirstType')
_SecondType = TypeVar('_SecondType')
_ThirdType = TypeVar('_ThirdType')
_UpdatedType = TypeVar('_UpdatedType')

_MappableKind = TypeVar('_MappableKind', bound=MappableN)


def map_(
    function: Callable[[_FirstType], _UpdatedType],
) -> Kinded[
    Callable[
        [KindN[_MappableKind, _FirstType, _SecondType, _ThirdType]],
        KindN[_MappableKind, _UpdatedType, _SecondType, _ThirdType],
    ]
]:
    """
    Lifts function to be wrapped in a container for better composition.

    In other words, it modifies the function's
    signature from:
    ``a -> b``
    to: ``Container[a] -> Container[b]``

    This is how it should be used:

    .. code:: python

        >>> from returns.io import IO
        >>> from returns.pointfree import map_

        >>> def example(argument: int) -> float:
        ...     return argument / 2

        >>> assert map_(example)(IO(1)) == IO(0.5)

    Note, that this function works for all containers with ``.map`` method.
    See :class:`returns.primitives.interfaces.mappable.MappableN` for more info.

    See also:
        - https://wiki.haskell.org/Lifting

    """

    @kinded
    def factory(
        container: KindN[_MappableKind, _FirstType, _SecondType, _ThirdType],
    ) -> KindN[_MappableKind, _UpdatedType, _SecondType, _ThirdType]:
        return container.map(function)

    return factory



from returns.result import Failure, Success


def test_success_value():
    """Ensures that value is fetch correctly from the Success."""
    bound = Success(5).value_or(None)

    assert bound == 5


def test_failure_value():
    """Ensures that value is fetch correctly from the Failure."""
    bound = Failure(1).value_or(default_value=None)

    assert bound is None



from collections.abc import Callable, Sequence
from typing import ClassVar, Final, Generic, TypeVar, final

from returns.primitives.types import Immutable

_Caps = TypeVar('_Caps')
_ReturnType = TypeVar('_ReturnType')
_TypeArgType1 = TypeVar('_TypeArgType1')
_TypeArgType2 = TypeVar('_TypeArgType2')
_TypeArgType3 = TypeVar('_TypeArgType3')

#: Special alias to define laws as functions even inside a class
law_definition = staticmethod

LAWS_ATTRIBUTE: Final = '_laws'


class Law(Immutable):
    """
    Base class for all laws. Does not have an attached signature.

    Should not be used directly.
    Use ``Law1``, ``Law2`` or ``Law3`` instead.
    """

    __slots__ = ('definition',)

    #: Function used to define this law.
    definition: Callable

    def __init__(self, function) -> None:
        """Saves function to the inner state."""
        object.__setattr__(self, 'definition', function)

    @final
    @property
    def name(self) -> str:
        """Returns a name of the given law. Basically a name of the function."""
        return self.definition.__name__


@final
class Law1(
    Law,
    Generic[_TypeArgType1, _ReturnType],
):
    """Law definition for functions with a single argument."""

    __slots__ = ()

    definition: Callable[['Law1', _TypeArgType1], _ReturnType]

    def __init__(
        self,
        function: Callable[[_TypeArgType1], _ReturnType],
    ) -> None:
        """Saves function of one argument to the inner state."""
        super().__init__(function)


@final
class Law2(
    Law,
    Generic[_TypeArgType1, _TypeArgType2, _ReturnType],
):
    """Law definition for functions with two arguments."""

    __slots__ = ()

    definition: Callable[['Law2', _TypeArgType1, _TypeArgType2], _ReturnType]

    def __init__(
        self,
        function: Callable[[_TypeArgType1, _TypeArgType2], _ReturnType],
    ) -> None:
        """Saves function of two arguments to the inner state."""
        super().__init__(function)


@final
class Law3(
    Law,
    Generic[_TypeArgType1, _TypeArgType2, _TypeArgType3, _ReturnType],
):
    """Law definition for functions with three argument."""

    __slots__ = ()

    definition: Callable[
        ['Law3', _TypeArgType1, _TypeArgType2, _TypeArgType3],
        _ReturnType,
    ]

    def __init__(
        self,
        function: Callable[
            [_TypeArgType1, _TypeArgType2, _TypeArgType3],
            _ReturnType,
        ],
    ) -> None:
        """Saves function of three arguments to the inner state."""
        super().__init__(function)


class Lawful(Generic[_Caps]):
    """
    Base class for all lawful classes.

    Allows to smartly collect all defined laws from all parent classes.
    """

    __slots__ = ()

    #: Some classes and interfaces might have laws, some might not have any.
    _laws: ClassVar[Sequence[Law]]

    @final  # noqa: WPS210
    @classmethod
    def laws(cls) -> dict[type['Lawful'], Sequence[Law]]:  # noqa: WPS210
        """
        Collects all laws from all parent classes.

        Algorithm:

        1. First, we collect all unique parents in ``__mro__``
        2. Then we get the laws definition from each of them
        3. Then we structure them in a ``type: its_laws`` way

        """
        seen = {
            f'{parent.__module__}.{parent.__qualname__}': parent
            for parent in cls.__mro__
        }

        laws = {}
        for klass in seen.values():
            current_laws = klass.__dict__.get(LAWS_ATTRIBUTE, ())
            if not current_laws:
                continue
            laws[klass] = current_laws
        return laws


class LawSpecDef:
    """Base class for all collection of laws aka LawSpecs."""

    __slots__ = ()



from typing import Never, TypeVar

from returns.interfaces import altable, mappable

_FirstType = TypeVar('_FirstType')
_SecondType = TypeVar('_SecondType')
_ThirdType = TypeVar('_ThirdType')


class BiMappableN(
    mappable.MappableN[_FirstType, _SecondType, _ThirdType],
    altable.AltableN[_FirstType, _SecondType, _ThirdType],
):
    """
    Allows to change both types of a container at the same time.

    Uses ``.map`` to change first type and ``.alt`` to change second type.

    See also:
        - https://typelevel.org/cats/typeclasses/bifunctor.html

    """

    __slots__ = ()


#: Type alias for kinds with two type arguments.
BiMappable2 = BiMappableN[_FirstType, _SecondType, Never]

#: Type alias for kinds with three type arguments.
BiMappable3 = BiMappableN[_FirstType, _SecondType, _ThirdType]



import pytest

from returns.future import FutureResult


def test_nonequality():
    """Ensures that containers can be compared."""
    assert FutureResult.from_value(1) != FutureResult.from_value(1)
    assert hash(FutureResult.from_value(1))


@pytest.mark.anyio
async def test_equality():
    """Ensures that containers are not compared to regular values."""
    assert await FutureResult.from_value(
        2,
    ) == await FutureResult.from_value(2)



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, Any, ClassVar, TypeAlias, TypeVar, final

from returns.context import NoDeps
from returns.interfaces.specific import reader_ioresult
from returns.io import IO, IOFailure, IOResult, IOSuccess
from returns.primitives.container import BaseContainer
from returns.primitives.hkt import Kind3, SupportsKind3, dekind
from returns.result import Result

if TYPE_CHECKING:
    from returns.context.requires_context import RequiresContext
    from returns.context.requires_context_result import RequiresContextResult

# Context:
_EnvType_contra = TypeVar('_EnvType_contra', contravariant=True)
_NewEnvType = TypeVar('_NewEnvType')

# Result:
_ValueType_co = TypeVar('_ValueType_co', covariant=True)
_NewValueType = TypeVar('_NewValueType')
_ErrorType = TypeVar('_ErrorType')
_NewErrorType = TypeVar('_NewErrorType')


@final
class RequiresContextIOResult(  # type: ignore[type-var]
    BaseContainer,
    SupportsKind3[
        'RequiresContextIOResult', _ValueType_co, _ErrorType, _EnvType_contra
    ],
    reader_ioresult.ReaderIOResultBasedN[
        _ValueType_co, _ErrorType, _EnvType_contra
    ],
):
    """
    The ``RequiresContextIOResult`` combinator.

    See :class:`returns.context.requires_context.RequiresContext`
    and :class:`returns.context.requires_context_result.RequiresContextResult`
    for more docs.

    This is just a handy wrapper around
    ``RequiresContext[IOResult[a, b], env]``
    which represents a context-dependent impure operation that might fail.

    It has several important differences from the regular ``Result`` classes.
    It does not have ``Success`` and ``Failure`` subclasses.
    Because, the computation is not yet performed.
    And we cannot know the type in advance.

    So, this is a thin wrapper, without any changes in logic.

    Why do we need this wrapper? That's just for better usability!

    .. code:: python

      >>> from returns.context import RequiresContext
      >>> from returns.io import IOSuccess, IOResult

      >>> def function(arg: int) -> IOResult[int, str]:
      ...      return IOSuccess(arg + 1)

      >>> # Without wrapper:
      >>> assert RequiresContext.from_value(IOSuccess(1)).map(
      ...     lambda ioresult: ioresult.bind(function),
      ... )(...) == IOSuccess(2)

      >>> # With wrapper:
      >>> assert RequiresContextIOResult.from_value(1).bind_ioresult(
      ...     function,
      ... )(...) == IOSuccess(2)

    This way ``RequiresContextIOResult`` allows to simply work with:

    - raw values and pure functions
    - ``RequiresContext`` values and pure functions returning it
    - ``RequiresContextResult`` values and pure functions returning it
    - ``Result`` and pure functions returning it
    - ``IOResult`` and functions returning it
    - other ``RequiresContextIOResult`` related functions and values

    This is a complex type for complex tasks!

    .. rubric:: Important implementation details

    Due it is meaning, ``RequiresContextIOResult``
    cannot have ``Success`` and ``Failure`` subclasses.

    We only have just one type. That's by design.

    Different converters are also not supported for this type.
    Use converters inside the ``RequiresContext`` context, not outside.

    See also:
        - https://dev.to/gcanti/getting-started-with-fp-ts-reader-1ie5
        - https://en.wikipedia.org/wiki/Lazy_evaluation
        - https://bit.ly/2R8l4WK
        - https://bit.ly/2RwP4fp

    """

    __slots__ = ()

    #: Inner value of `RequiresContext`
    #: is just a function that returns `IOResult`.
    #: This field has an extra 'RequiresContext' just because `mypy` needs it.
    _inner_value: Callable[
        [_EnvType_contra], IOResult[_ValueType_co, _ErrorType]
    ]

    #: A convenient placeholder to call methods created by `.from_value()`.
    no_args: ClassVar[NoDeps] = object()

    def __init__(
        self,
        inner_value: Callable[
            [_EnvType_contra], IOResult[_ValueType_co, _ErrorType]
        ],
    ) -> None:
        """
        Public constructor for this type. Also required for typing.

        Only allows functions of kind ``* -> *``
        and returning :class:`returns.result.Result` instances.

        .. code:: python

          >>> from returns.context import RequiresContextIOResult
          >>> from returns.io import IOSuccess
          >>> str(RequiresContextIOResult(lambda deps: IOSuccess(deps + 1)))
          '<RequiresContextIOResult: <function <lambda> at ...>>'

        """
        super().__init__(inner_value)

    def __call__(
        self, deps: _EnvType_contra
    ) -> IOResult[_ValueType_co, _ErrorType]:
        """
        Evaluates the wrapped function.

        .. code:: python

          >>> from returns.context import RequiresContextIOResult
          >>> from returns.io import IOSuccess

          >>> def first(lg: bool) -> RequiresContextIOResult[int, str, float]:
          ...     # `deps` has `float` type here:
          ...     return RequiresContextIOResult(
          ...         lambda deps: IOSuccess(deps if lg else -deps),
          ...     )

          >>> instance = first(False)
          >>> assert instance(3.5) == IOSuccess(-3.5)

        In other things, it is a regular Python magic method.

        """
        return self._inner_value(deps)

    def swap(
        self,
    ) -> RequiresContextIOResult[_ErrorType, _ValueType_co, _EnvType_contra]:
        """
        Swaps value and error types.

        So, values become errors and errors become values.
        It is useful when you have to work with errors a lot.
        And since we have a lot of ``.bind_`` related methods
        and only a single ``.lash`` - it is easier to work with values.

        .. code:: python

          >>> from returns.context import RequiresContextIOResult
          >>> from returns.io import IOSuccess, IOFailure

          >>> success = RequiresContextIOResult.from_value(1)
          >>> failure = RequiresContextIOResult.from_failure(1)

          >>> assert success.swap()(...) == IOFailure(1)
          >>> assert failure.swap()(...) == IOSuccess(1)

        """
        return RequiresContextIOResult(lambda deps: self(deps).swap())

    def map(
        self,
        function: Callable[[_ValueType_co], _NewValueType],
    ) -> RequiresContextIOResult[_NewValueType, _ErrorType, _EnvType_contra]:
        """
        Composes successful container with a pure function.

        .. code:: python

          >>> from returns.context import RequiresContextIOResult
          >>> from returns.io import IOSuccess, IOFailure

          >>> assert RequiresContextIOResult.from_value(1).map(
          ...     lambda x: x + 1,
          ... )(...) == IOSuccess(2)

          >>> assert RequiresContextIOResult.from_failure(1).map(
          ...     lambda x: x + 1,
          ... )(...) == IOFailure(1)

        """
        return RequiresContextIOResult(lambda deps: self(deps).map(function))

    def apply(
        self,
        container: Kind3[
            RequiresContextIOResult,
            Callable[[_ValueType_co], _NewValueType],
            _ErrorType,
            _EnvType_contra,
        ],
    ) -> RequiresContextIOResult[_NewValueType, _ErrorType, _EnvType_contra]:
        """
        Calls a wrapped function in a container on this container.

        .. code:: python

          >>> from returns.context import RequiresContextIOResult
          >>> from returns.io import IOSuccess, IOFailure

          >>> def transform(arg: str) -> str:
          ...     return arg + 'b'

          >>> assert RequiresContextIOResult.from_value('a').apply(
          ...    RequiresContextIOResult.from_value(transform),
          ... )(...) == IOSuccess('ab')

          >>> assert RequiresContextIOResult.from_value('a').apply(
          ...    RequiresContextIOResult.from_failure(1),
          ... )(...) == IOFailure(1)

          >>> assert RequiresContextIOResult.from_failure('a').apply(
          ...    RequiresContextIOResult.from_value(transform),
          ... )(...) == IOFailure('a')

          >>> assert RequiresContextIOResult.from_failure('a').apply(
          ...    RequiresContextIOResult.from_failure('b'),
          ... )(...) == IOFailure('a')

        """
        return RequiresContextIOResult(
            lambda deps: self(deps).apply(dekind(container)(deps)),
        )

    def bind(
        self,
        function: Callable[
            [_ValueType_co],
            Kind3[
                RequiresContextIOResult,
                _NewValueType,
                _ErrorType,
                _EnvType_contra,
            ],
        ],
    ) -> RequiresContextIOResult[_NewValueType, _ErrorType, _EnvType_contra]:
        """
        Composes this container with a function returning the same type.

        .. code:: python

          >>> from returns.context import RequiresContextIOResult
          >>> from returns.io import IOSuccess, IOFailure

          >>> def first(lg: bool) -> RequiresContextIOResult[int, int, float]:
          ...     # `deps` has `float` type here:
          ...     return RequiresContextIOResult(
          ...         lambda deps: IOSuccess(deps) if lg else IOFailure(-deps),
          ...     )

          >>> def second(
          ...     number: int,
          ... ) -> RequiresContextIOResult[str, int, float]:
          ...     # `deps` has `float` type here:
          ...     return RequiresContextIOResult(
          ...         lambda deps: IOSuccess('>=' if number >= deps else '<'),
          ...     )

          >>> assert first(True).bind(second)(1) == IOSuccess('>=')
          >>> assert first(False).bind(second)(2) == IOFailure(-2)

        """
        return RequiresContextIOResult(
            lambda deps: self(deps).bind(
                lambda inner: dekind(  # type: ignore[misc]
                    function(inner),
                )(deps),
            ),
        )

    #: Alias for `bind_context_ioresult` method, it is the same as `bind` here.
    bind_context_ioresult = bind

    def bind_result(
        self,
        function: Callable[[_ValueType_co], Result[_NewValueType, _ErrorType]],
    ) -> RequiresContextIOResult[_NewValueType, _ErrorType, _EnvType_contra]:
        """
        Binds ``Result`` returning function to the current container.

        .. code:: python

          >>> from returns.context import RequiresContextIOResult
          >>> from returns.result import Failure, Result, Success
          >>> from returns.io import IOSuccess, IOFailure

          >>> def function(num: int) -> Result[int, str]:
          ...     return Success(num + 1) if num > 0 else Failure('<0')

          >>> assert RequiresContextIOResult.from_value(1).bind_result(
          ...     function,
          ... )(RequiresContextIOResult.no_args) == IOSuccess(2)

          >>> assert RequiresContextIOResult.from_value(0).bind_result(
          ...     function,
          ... )(RequiresContextIOResult.no_args) == IOFailure('<0')

          >>> assert RequiresContextIOResult.from_failure(':(').bind_result(
          ...     function,
          ... )(RequiresContextIOResult.no_args) == IOFailure(':(')

        """
        return RequiresContextIOResult(
            lambda deps: self(deps).bind_result(function),
        )

    def bind_context(
        self,
        function: Callable[
            [_ValueType_co],
            RequiresContext[_NewValueType, _EnvType_contra],
        ],
    ) -> RequiresContextIOResult[_NewValueType, _ErrorType, _EnvType_contra]:
        """
        Binds ``RequiresContext`` returning function to current container.

        .. code:: python

          >>> from returns.context import RequiresContext
          >>> from returns.io import IOSuccess, IOFailure

          >>> def function(arg: int) -> RequiresContext[int, str]:
          ...     return RequiresContext(lambda deps: len(deps) + arg)

          >>> assert function(2)('abc') == 5

          >>> assert RequiresContextIOResult.from_value(2).bind_context(
          ...     function,
          ... )('abc') == IOSuccess(5)

          >>> assert RequiresContextIOResult.from_failure(2).bind_context(
          ...     function,
          ... )('abc') == IOFailure(2)

        """
        return RequiresContextIOResult(
            lambda deps: self(deps).map(
                lambda inner: function(inner)(deps),  # type: ignore[misc]
            ),
        )

    def bind_context_result(
        self,
        function: Callable[
            [_ValueType_co],
            RequiresContextResult[_NewValueType, _ErrorType, _EnvType_contra],
        ],
    ) -> RequiresContextIOResult[_NewValueType, _ErrorType, _EnvType_contra]:
        """
        Binds ``RequiresContextResult`` returning function to the current one.

        .. code:: python

          >>> from returns.context import RequiresContextResult
          >>> from returns.io import IOSuccess, IOFailure
          >>> from returns.result import Success, Failure

          >>> def function(arg: int) -> RequiresContextResult[int, int, str]:
          ...     if arg > 0:
          ...         return RequiresContextResult(
          ...             lambda deps: Success(len(deps) + arg),
          ...         )
          ...     return RequiresContextResult(
          ...         lambda deps: Failure(len(deps) + arg),
          ...     )

          >>> assert function(2)('abc') == Success(5)
          >>> assert function(-1)('abc') == Failure(2)

          >>> assert RequiresContextIOResult.from_value(
          ...    2,
          ... ).bind_context_result(
          ...     function,
          ... )('abc') == IOSuccess(5)

          >>> assert RequiresContextIOResult.from_value(
          ...    -1,
          ... ).bind_context_result(
          ...     function,
          ... )('abc') == IOFailure(2)

          >>> assert RequiresContextIOResult.from_failure(
          ...    2,
          ... ).bind_context_result(
          ...     function,
          ... )('abc') == IOFailure(2)

        """
        return RequiresContextIOResult(
            lambda deps: self(deps).bind_result(
                lambda inner: function(inner)(deps),  # type: ignore[misc]
            ),
        )

    def bind_io(
        self,
        function: Callable[[_ValueType_co], IO[_NewValueType]],
    ) -> RequiresContextIOResult[_NewValueType, _ErrorType, _EnvType_contra]:
        """
        Binds ``IO`` returning function to the current container.

        .. code:: python

          >>> from returns.context import RequiresContextIOResult
          >>> from returns.io import IO, IOSuccess, IOFailure

          >>> def function(number: int) -> IO[str]:
          ...     return IO(str(number))

          >>> assert RequiresContextIOResult.from_value(1).bind_io(
          ...     function,
          ... )(RequiresContextIOResult.no_args) == IOSuccess('1')

          >>> assert RequiresContextIOResult.from_failure(1).bind_io(
          ...     function,
          ... )(RequiresContextIOResult.no_args) == IOFailure(1)

        """
        return RequiresContextIOResult(
            lambda deps: self(deps).bind_io(function),
        )

    def bind_ioresult(
        self,
        function: Callable[
            [_ValueType_co], IOResult[_NewValueType, _ErrorType]
        ],
    ) -> RequiresContextIOResult[_NewValueType, _ErrorType, _EnvType_contra]:
        """
        Binds ``IOResult`` returning function to the current container.

        .. code:: python

          >>> from returns.context import RequiresContextIOResult
          >>> from returns.io import IOResult, IOSuccess, IOFailure

          >>> def function(num: int) -> IOResult[int, str]:
          ...     return IOSuccess(num + 1) if num > 0 else IOFailure('<0')

          >>> assert RequiresContextIOResult.from_value(1).bind_ioresult(
          ...     function,
          ... )(RequiresContextIOResult.no_args) == IOSuccess(2)

          >>> assert RequiresContextIOResult.from_value(0).bind_ioresult(
          ...     function,
          ... )(RequiresContextIOResult.no_args) == IOFailure('<0')

          >>> assert RequiresContextIOResult.from_failure(':(').bind_ioresult(
          ...     function,
          ... )(RequiresContextIOResult.no_args) == IOFailure(':(')

        """
        return RequiresContextIOResult(
            lambda deps: self(deps).bind(function),
        )

    def alt(
        self,
        function: Callable[[_ErrorType], _NewErrorType],
    ) -> RequiresContextIOResult[_ValueType_co, _NewErrorType, _EnvType_contra]:
        """
        Composes failed container with a pure function.

        .. code:: python

          >>> from returns.context import RequiresContextIOResult
          >>> from returns.io import IOSuccess, IOFailure

          >>> assert RequiresContextIOResult.from_value(1).alt(
          ...     lambda x: x + 1,
          ... )(...) == IOSuccess(1)

          >>> assert RequiresContextIOResult.from_failure(1).alt(
          ...     lambda x: x + 1,
          ... )(...) == IOFailure(2)

        """
        return RequiresContextIOResult(lambda deps: self(deps).alt(function))

    def lash(
        self,
        function: Callable[
            [_ErrorType],
            Kind3[
                RequiresContextIOResult,
                _ValueType_co,
                _NewErrorType,
                _EnvType_contra,
            ],
        ],
    ) -> RequiresContextIOResult[_ValueType_co, _NewErrorType, _EnvType_contra]:
        """
        Composes this container with a function returning the same type.

        .. code:: python

          >>> from returns.context import RequiresContextIOResult
          >>> from returns.io import IOSuccess, IOFailure

          >>> def lashable(
          ...     arg: str,
          ... ) -> RequiresContextIOResult[str, str, str]:
          ...      if len(arg) > 1:
          ...          return RequiresContextIOResult(
          ...              lambda deps: IOSuccess(deps + arg),
          ...          )
          ...      return RequiresContextIOResult(
          ...          lambda deps: IOFailure(arg + deps),
          ...      )

          >>> assert RequiresContextIOResult.from_value('a').lash(
          ...     lashable,
          ... )('c') == IOSuccess('a')
          >>> assert RequiresContextIOResult.from_failure('a').lash(
          ...     lashable,
          ... )('c') == IOFailure('ac')
          >>> assert RequiresContextIOResult.from_failure('aa').lash(
          ...     lashable,
          ... )('b') == IOSuccess('baa')

        """
        return RequiresContextIOResult(
            lambda deps: self(deps).lash(
                lambda inner: function(inner)(deps),  # type: ignore
            ),
        )

    def compose_result(
        self,
        function: Callable[
            [Result[_ValueType_co, _ErrorType]],
            Kind3[
                RequiresContextIOResult,
                _NewValueType,
                _ErrorType,
                _EnvType_contra,
            ],
        ],
    ) -> RequiresContextIOResult[_NewValueType, _ErrorType, _EnvType_contra]:
        """
        Composes inner ``Result`` with ``ReaderIOResult`` returning function.

        Can be useful when you need an access to both states of the result.

        .. code:: python

          >>> from returns.context import ReaderIOResult, NoDeps
          >>> from returns.io import IOSuccess, IOFailure
          >>> from returns.result import Result

          >>> def count(
          ...    container: Result[int, int],
          ... ) -> ReaderIOResult[int, int, NoDeps]:
          ...     return ReaderIOResult.from_result(
          ...         container.map(lambda x: x + 1).alt(abs),
          ...     )

          >>> success = ReaderIOResult.from_value(1)
          >>> failure = ReaderIOResult.from_failure(-1)
          >>> assert success.compose_result(count)(...) == IOSuccess(2)
          >>> assert failure.compose_result(count)(...) == IOFailure(1)

        """
        return RequiresContextIOResult(
            lambda deps: dekind(
                function(self(deps)._inner_value),  # noqa: SLF001
            )(deps),
        )

    def modify_env(
        self,
        function: Callable[[_NewEnvType], _EnvType_contra],
    ) -> RequiresContextIOResult[_ValueType_co, _ErrorType, _NewEnvType]:
        """
        Allows to modify the environment type.

        .. code:: python

          >>> from returns.context import RequiresContextIOResultE
          >>> from returns.io import IOSuccess, impure_safe

          >>> def div(arg: int) -> RequiresContextIOResultE[float, int]:
          ...     return RequiresContextIOResultE(
          ...         impure_safe(lambda deps: arg / deps),
          ...     )

          >>> assert div(3).modify_env(int)('2') == IOSuccess(1.5)
          >>> assert div(3).modify_env(int)('0').failure()

        """
        return RequiresContextIOResult(lambda deps: self(function(deps)))

    @classmethod
    def ask(
        cls,
    ) -> RequiresContextIOResult[_EnvType_contra, _ErrorType, _EnvType_contra]:
        """
        Is used to get the current dependencies inside the call stack.

        Similar to :meth:`returns.context.requires_context.RequiresContext.ask`,
        but returns ``IOResult`` instead of a regular value.

        Please, refer to the docs there to learn how to use it.

        One important note that is worth duplicating here:
        you might need to provide ``_EnvType_contra`` explicitly,
        so ``mypy`` will know about it statically.

        .. code:: python

          >>> from returns.context import RequiresContextIOResultE
          >>> from returns.io import IOSuccess
          >>> assert RequiresContextIOResultE[int, int].ask().map(
          ...     str,
          ... )(1) == IOSuccess('1')

        """
        return RequiresContextIOResult(IOSuccess)

    @classmethod
    def from_result(
        cls,
        inner_value: Result[_NewValueType, _NewErrorType],
    ) -> RequiresContextIOResult[_NewValueType, _NewErrorType, NoDeps]:
        """
        Creates new container with ``Result`` as a unit value.

        .. code:: python

          >>> from returns.context import RequiresContextIOResult
          >>> from returns.result import Success, Failure
          >>> from returns.io import IOSuccess, IOFailure
          >>> deps = RequiresContextIOResult.no_args

          >>> assert RequiresContextIOResult.from_result(
          ...    Success(1),
          ... )(deps) == IOSuccess(1)

          >>> assert RequiresContextIOResult.from_result(
          ...    Failure(1),
          ... )(deps) == IOFailure(1)

        """
        return RequiresContextIOResult(
            lambda _: IOResult.from_result(inner_value),
        )

    @classmethod
    def from_io(
        cls,
        inner_value: IO[_NewValueType],
    ) -> RequiresContextIOResult[_NewValueType, Any, NoDeps]:
        """
        Creates new container from successful ``IO`` value.

        .. code:: python

          >>> from returns.io import IO, IOSuccess
          >>> from returns.context import RequiresContextIOResult

          >>> assert RequiresContextIOResult.from_io(IO(1))(
          ...     RequiresContextIOResult.no_args,
          ... ) == IOSuccess(1)

        """
        return RequiresContextIOResult(
            lambda deps: IOResult.from_io(inner_value),
        )

    @classmethod
    def from_failed_io(
        cls,
        inner_value: IO[_NewErrorType],
    ) -> RequiresContextIOResult[Any, _NewErrorType, NoDeps]:
        """
        Creates a new container from failed ``IO`` value.

        .. code:: python

          >>> from returns.io import IO, IOFailure
          >>> from returns.context import RequiresContextIOResult

          >>> assert RequiresContextIOResult.from_failed_io(IO(1))(
          ...     RequiresContextIOResult.no_args,
          ... ) == IOFailure(1)

        """
        return RequiresContextIOResult(
            lambda deps: IOResult.from_failed_io(inner_value),
        )

    @classmethod
    def from_ioresult(
        cls,
        inner_value: IOResult[_NewValueType, _NewErrorType],
    ) -> RequiresContextIOResult[_NewValueType, _NewErrorType, NoDeps]:
        """
        Creates new container with ``IOResult`` as a unit value.

        .. code:: python

          >>> from returns.context import RequiresContextIOResult
          >>> from returns.io import IOSuccess, IOFailure
          >>> deps = RequiresContextIOResult.no_args

          >>> assert RequiresContextIOResult.from_ioresult(
          ...    IOSuccess(1),
          ... )(deps) == IOSuccess(1)

          >>> assert RequiresContextIOResult.from_ioresult(
          ...    IOFailure(1),
          ... )(deps) == IOFailure(1)

        """
        return RequiresContextIOResult(lambda _: inner_value)

    @classmethod
    def from_ioresult_context(
        cls,
        inner_value: ReaderIOResult[_NewValueType, _NewErrorType, _NewEnvType],
    ) -> ReaderIOResult[_NewValueType, _NewErrorType, _NewEnvType]:
        """
        Creates new container with ``ReaderIOResult`` as a unit value.

        .. code:: python

          >>> from returns.context import RequiresContextIOResult
          >>> from returns.io import IOSuccess, IOFailure

          >>> assert RequiresContextIOResult.from_ioresult_context(
          ...    RequiresContextIOResult.from_value(1),
          ... )(...) == IOSuccess(1)

          >>> assert RequiresContextIOResult.from_ioresult_context(
          ...    RequiresContextIOResult.from_failure(1),
          ... )(...) == IOFailure(1)

        """
        return inner_value

    @classmethod
    def from_typecast(
        cls,
        inner_value: RequiresContext[
            IOResult[_NewValueType, _NewErrorType],
            _EnvType_contra,
        ],
    ) -> RequiresContextIOResult[_NewValueType, _NewErrorType, _EnvType_contra]:
        """
        You might end up with ``RequiresContext[IOResult]`` as a value.

        This method is designed to turn it into ``RequiresContextIOResult``.
        It will save all the typing information.

        It is just more useful!

        .. code:: python

          >>> from returns.context import RequiresContext
          >>> from returns.io import IOSuccess, IOFailure

          >>> assert RequiresContextIOResult.from_typecast(
          ...     RequiresContext.from_value(IOSuccess(1)),
          ... )(RequiresContextIOResult.no_args) == IOSuccess(1)

          >>> assert RequiresContextIOResult.from_typecast(
          ...     RequiresContext.from_value(IOFailure(1)),
          ... )(RequiresContextIOResult.no_args) == IOFailure(1)

        """
        return RequiresContextIOResult(inner_value)

    @classmethod
    def from_context(
        cls,
        inner_value: RequiresContext[_NewValueType, _NewEnvType],
    ) -> RequiresContextIOResult[_NewValueType, Any, _NewEnvType]:
        """
        Creates new container from ``RequiresContext`` as a success unit.

        .. code:: python

          >>> from returns.context import RequiresContext
          >>> from returns.io import IOSuccess

          >>> assert RequiresContextIOResult.from_context(
          ...     RequiresContext.from_value(1),
          ... )(...) == IOSuccess(1)

        """
        return RequiresContextIOResult(
            lambda deps: IOSuccess(inner_value(deps)),
        )

    @classmethod
    def from_failed_context(
        cls,
        inner_value: RequiresContext[_NewValueType, _NewEnvType],
    ) -> RequiresContextIOResult[Any, _NewValueType, _NewEnvType]:
        """
        Creates new container from ``RequiresContext`` as a failure unit.

        .. code:: python

          >>> from returns.context import RequiresContext
          >>> from returns.io import IOFailure

          >>> assert RequiresContextIOResult.from_failed_context(
          ...     RequiresContext.from_value(1),
          ... )(...) == IOFailure(1)

        """
        return RequiresContextIOResult(
            lambda deps: IOFailure(inner_value(deps)),
        )

    @classmethod
    def from_result_context(
        cls,
        inner_value: RequiresContextResult[
            _NewValueType,
            _NewErrorType,
            _NewEnvType,
        ],
    ) -> RequiresContextIOResult[_NewValueType, _NewErrorType, _NewEnvType]:
        """
        Creates new container from ``RequiresContextResult`` as a unit value.

        .. code:: python

          >>> from returns.context import RequiresContextResult
          >>> from returns.io import IOSuccess, IOFailure

          >>> assert RequiresContextIOResult.from_result_context(
          ...     RequiresContextResult.from_value(1),
          ... )(...) == IOSuccess(1)

          >>> assert RequiresContextIOResult.from_result_context(
          ...     RequiresContextResult.from_failure(1),
          ... )(...) == IOFailure(1)

        """
        return RequiresContextIOResult(
            lambda deps: IOResult.from_result(inner_value(deps)),
        )

    @classmethod
    def from_value(
        cls,
        inner_value: _NewValueType,
    ) -> RequiresContextIOResult[_NewValueType, Any, NoDeps]:
        """
        Creates new container with ``IOSuccess(inner_value)`` as a unit value.

        .. code:: python

          >>> from returns.context import RequiresContextIOResult
          >>> from returns.io import IOSuccess

          >>> assert RequiresContextIOResult.from_value(1)(
          ...    RequiresContextIOResult.no_args,
          ... ) == IOSuccess(1)

        """
        return RequiresContextIOResult(lambda _: IOSuccess(inner_value))

    @classmethod
    def from_failure(
        cls,
        inner_value: _NewErrorType,
    ) -> RequiresContextIOResult[Any, _NewErrorType, NoDeps]:
        """
        Creates new container with ``IOFailure(inner_value)`` as a unit value.

        .. code:: python

          >>> from returns.context import RequiresContextIOResult
          >>> from returns.io import IOFailure

          >>> assert RequiresContextIOResult.from_failure(1)(
          ...     RequiresContextIOResult.no_args,
          ... ) == IOFailure(1)

        """
        return RequiresContextIOResult(lambda _: IOFailure(inner_value))


# Aliases:

#: Alias for a popular case when ``Result`` has ``Exception`` as error type.
RequiresContextIOResultE: TypeAlias = RequiresContextIOResult[
    _ValueType_co,
    Exception,
    _EnvType_contra,
]

#: Alias to save you some typing. Uses original name from Haskell.
ReaderIOResult: TypeAlias = RequiresContextIOResult

#: Alias to save you some typing. Uses ``Exception`` as error type.
ReaderIOResultE: TypeAlias = RequiresContextIOResult[
    _ValueType_co,
    Exception,
    _EnvType_contra,
]



from collections.abc import Iterator
from itertools import starmap
from typing import Final, final

from mypy.nodes import ARG_STAR, ARG_STAR2
from mypy.plugin import FunctionContext
from mypy.types import (
    CallableType,
    FunctionLike,
    Instance,
    Overloaded,
    ProperType,
    TypeType,
    get_proper_type,
)

from returns.contrib.mypy._structures.args import FuncArg
from returns.contrib.mypy._typeops.analtype import (
    analyze_call,
    safe_translate_to_function,
)
from returns.contrib.mypy._typeops.inference import CallableInference
from returns.contrib.mypy._typeops.transform_callable import (
    Functions,
    Intermediate,
    detach_callable,
    proper_type,
)

_SUPPORTED_TYPES: Final = (
    CallableType,
    Instance,
    TypeType,
    Overloaded,
)


def analyze(ctx: FunctionContext) -> ProperType:
    """
    This hook is used to make typed curring a thing in `returns` project.

    This plugin is a temporary solution to the problem.
    It should be later replaced with the official way of doing things.
    One day functions will have better API and we plan
    to submit this plugin into ``mypy`` core plugins,
    so it would not be required.

    Internally we just reduce the original function's argument count.
    And drop some of them from function's signature.
    """
    default_return = get_proper_type(ctx.default_return_type)
    if not isinstance(default_return, CallableType):
        return default_return

    function_def = get_proper_type(ctx.arg_types[0][0])
    func_args = _AppliedArgs(ctx)

    if len(list(filter(len, ctx.arg_types))) == 1:
        return function_def  # this means, that `partial(func)` is called
    if not isinstance(function_def, _SUPPORTED_TYPES):
        return default_return
    if isinstance(function_def, Instance | TypeType):
        # We force `Instance` and similar types to coercse to callable:
        function_def = func_args.get_callable_from_context()

    is_valid, applied_args = func_args.build_from_context()
    if not isinstance(function_def, CallableType | Overloaded) or not is_valid:
        return default_return

    return _PartialFunctionReducer(
        default_return,
        function_def,
        applied_args,
        ctx,
    ).new_partial()


@final
class _PartialFunctionReducer:
    """
    Helper object to work with curring.

    Here's a quick overview of things that is going on inside:

    1. Firstly we create intermediate callable that represents a subset
       of argument that are passed with the ``curry`` call
    2. Then, we run typechecking on this intermediate function
       and passed arguments to make sure that everything is correct
    3. Then, we subtract intermediate arguments from the passed function
    4. Finally we run type substitution on newly created final function
       to replace generic vars we already know to make sure
       that everything still works and the number of type vars is reduced

    This plugin requires several things:

    - One should now how ``ExpressionChecker`` from ``mypy`` works
    - What ``FunctionLike`` is
    - How kinds work in type checking
    - What ``map_actuals_to_formals`` is
    - How constraints work

    That's not an easy plugin to work with.
    """

    def __init__(
        self,
        default_return_type: FunctionLike,
        original: FunctionLike,
        applied_args: list[FuncArg],
        ctx: FunctionContext,
    ) -> None:
        """
        Saving the things we need.

        Args:
            default_return_type: default callable type got by ``mypy``.
            original: passed function to be curried.
            applied_args: arguments that are already provided in the definition.
            ctx: plugin hook context provided by ``mypy``.

        """
        self._default_return_type = default_return_type
        self._original = original
        self._applied_args = applied_args
        self._ctx = ctx

        self._case_functions: list[CallableType] = []
        self._fallbacks: list[CallableType] = []

    def new_partial(self) -> ProperType:
        """
        Creates new partial functions.

        Splits passed functions into ``case_function``
        where each overloaded spec is processed inducidually.
        Then we combine everything back together removing unfit parts.
        """
        for case_function in self._original.items:
            fallback, intermediate = self._create_intermediate(case_function)
            self._fallbacks.append(fallback)

            if intermediate:
                partial = self._create_partial_case(
                    case_function,
                    intermediate,
                    fallback,
                )
                self._case_functions.append(partial)
        return self._create_new_partial()

    def _create_intermediate(
        self,
        case_function: CallableType,
    ) -> tuple[CallableType, CallableType | None]:
        intermediate = Intermediate(case_function).with_applied_args(
            self._applied_args,
        )
        return intermediate, analyze_call(
            intermediate,
            self._applied_args,
            self._ctx,
            show_errors=False,
        )

    def _create_partial_case(
        self,
        case_function: CallableType,
        intermediate: CallableType,
        fallback: CallableType,
    ) -> CallableType:
        partial = CallableInference(
            Functions(case_function, intermediate).diff(),
            self._ctx,
            fallback=fallback,
        ).from_usage(self._applied_args)

        if case_function.is_generic():
            # We can deal with really different `case_function` over here.
            # The first one is regular `generic` function
            # that has variables and typevars in its spec.
            # In this case, we process `partial` the same way.
            # It should be generic also.
            #
            # The second possible type of `case_function` is pseudo-generic.
            # These are functions that contain typevars in its spec,
            # but variables are empty.
            # Probably these functions are already used in a generic context.
            # So, we ignore them and do not add variables back.
            #
            # Regular functions are also untouched by this.
            return detach_callable(partial)
        return partial.copy_modified(variables=[])

    def _create_new_partial(self) -> ProperType:
        """
        Creates a new partial function-like from set of callables.

        We also need fallbacks here, because sometimes
        there are no possible ways to create at least a single partial case.
        In this scenario we analyze the set of fallbacks
        and tell user what went wrong.
        """
        if not self._case_functions:
            analyze_call(
                proper_type(self._fallbacks),
                self._applied_args,
                self._ctx,
                show_errors=True,
            )
            return self._default_return_type
        return proper_type(self._case_functions)


@final
class _AppliedArgs:
    """Builds applied args that were partially applied."""

    def __init__(self, function_ctx: FunctionContext) -> None:
        """
        We need the function default context.

        The first arguments of ``partial`` is skipped:
        it is the applied function itself.
        """
        self._function_ctx = function_ctx
        self._parts = zip(
            self._function_ctx.arg_names[1:],
            self._function_ctx.arg_types[1:],
            self._function_ctx.arg_kinds[1:],
            strict=False,
        )

    def get_callable_from_context(self) -> ProperType:
        """Returns callable type from the context."""
        return get_proper_type(
            safe_translate_to_function(
                self._function_ctx.arg_types[0][0],
                self._function_ctx,
            )
        )

    def build_from_context(self) -> tuple[bool, list[FuncArg]]:
        """
        Builds handy arguments structures from the context.

        Some usages might be invalid,
        because we cannot really infer some arguments.

        .. code:: python

            partial(some, *args)
            partial(other, **kwargs)

        Here ``*args`` and ``**kwargs`` can be literally anything!
        In these cases we fallback to the default return type.
        """
        applied_args = []
        for names, types, kinds in self._parts:
            for arg in self._generate_applied_args(
                zip(names, types, kinds, strict=False)
            ):
                if arg.kind in {ARG_STAR, ARG_STAR2}:
                    # We cannot really work with `*args`, `**kwargs`.
                    return False, []

                applied_args.append(arg)
        return True, applied_args

    def _generate_applied_args(self, arg_parts) -> Iterator[FuncArg]:
        yield from starmap(FuncArg, arg_parts)



import pytest

from returns.contrib.pytest import ReturnsAsserts
from returns.io import IOFailure, IOSuccess
from returns.result import Failure, Success, safe


def _create_container_function(container_type, container_value):
    return container_type(container_value)


def _create_container_function_intermediate(container_type, container_value):
    return _create_container_function(  # type: ignore
        container_type,
        container_value,
    )


@safe
def _safe_decorated_function(*, return_failure: bool = False):
    if return_failure:
        raise ValueError('Function failed')


@pytest.mark.parametrize(
    'container_type',
    [  # noqa: WPS118
        Success,
        Failure,
        IOSuccess,
        IOFailure,
    ],
)
def test_assert_trace1(container_type, returns: ReturnsAsserts):
    """Test if our plugin will identify the container creation correctly."""
    with returns.assert_trace(container_type, _create_container_function):
        _create_container_function(container_type, 1)  # type: ignore


@pytest.mark.parametrize(
    'container_type',
    [  # noqa: WPS118
        Success,
        Failure,
        IOSuccess,
        IOFailure,
    ],
)
def test_assert_trace2(container_type, returns: ReturnsAsserts):
    """Test if our plugin will identify the container creation correctly."""
    with returns.assert_trace(container_type, _create_container_function):
        _create_container_function_intermediate(  # type: ignore
            container_type,
            1,
        )


@pytest.mark.parametrize(
    ('desired_type', 'wrong_type'),
    [
        (Success, Failure),
        (Failure, Success),
        (IOSuccess, IOFailure),
        (IOFailure, IOSuccess),
    ],
)
def test_failed_assert_trace1(
    desired_type,
    wrong_type,
    returns: ReturnsAsserts,
):
    """Test if our plugin will identify the container was not created."""
    with (
        pytest.raises(pytest.fail.Exception),
        returns.assert_trace(desired_type, _create_container_function),
    ):
        _create_container_function(wrong_type, 1)  # type: ignore


@pytest.mark.parametrize(
    ('desired_type', 'wrong_type'),
    [
        (Success, Failure),
        (Failure, Success),
        (IOSuccess, IOFailure),
        (IOFailure, IOSuccess),
    ],
)
def test_failed_assert_trace2(
    desired_type,
    wrong_type,
    returns: ReturnsAsserts,
):
    """Test if our plugin will identify the container was not created."""
    with (
        pytest.raises(pytest.fail.Exception),
        returns.assert_trace(desired_type, _create_container_function),
    ):
        _create_container_function_intermediate(  # type: ignore
            wrong_type,
            1,
        )


@pytest.mark.parametrize(
    'container_type',
    [  # noqa: WPS118
        Success,
        Failure,
    ],
)
def test_safe_decorated_assert(container_type, returns: ReturnsAsserts):
    """Test if our plugin will catch containers from @safe-wrapped functions."""
    with returns.assert_trace(container_type, _safe_decorated_function):
        _safe_decorated_function(return_failure=container_type is Failure)



from returns.context import (
    Reader,
    ReaderFutureResult,
    ReaderIOResult,
    ReaderResult,
)
from returns.contrib.hypothesis.laws import check_all_laws
from returns.future import Future, FutureResult
from returns.io import IO, IOResult
from returns.maybe import Maybe
from returns.result import Result

check_all_laws(Maybe)
check_all_laws(Result)

check_all_laws(IO)
check_all_laws(IOResult)

check_all_laws(Future)
check_all_laws(FutureResult)

check_all_laws(Reader)
check_all_laws(ReaderResult)
check_all_laws(ReaderIOResult)
check_all_laws(ReaderFutureResult)



from collections.abc import Callable
from functools import partial as _partial
from functools import wraps
from inspect import BoundArguments, Signature
from typing import Any, TypeAlias, TypeVar

_ReturnType = TypeVar('_ReturnType')


def partial(
    func: Callable[..., _ReturnType],
    *args: Any,
    **kwargs: Any,
) -> Callable[..., _ReturnType]:
    """
    Typed partial application.

    It is just a ``functools.partial`` wrapper with better typing support.

    We use a custom ``mypy`` plugin to make sure types are correct.
    Otherwise, it is currently impossible to properly type this function.

    .. code:: python

      >>> from returns.curry import partial

      >>> def sum_two_numbers(first: int, second: int) -> int:
      ...     return first + second

      >>> sum_with_ten = partial(sum_two_numbers, 10)
      >>> assert sum_with_ten(2) == 12
      >>> assert sum_with_ten(-5) == 5

    See also:
        - https://docs.python.org/3/library/functools.html#functools.partial

    """
    return _partial(func, *args, **kwargs)


def curry(function: Callable[..., _ReturnType]) -> Callable[..., _ReturnType]:
    """
    Typed currying decorator.

    Currying is a conception from functional languages that does partial
    applying. That means that if we pass one argument in a function that
    gets 2 or more arguments, we'll get a new function that remembers all
    previously passed arguments. Then we can pass remaining arguments, and
    the function will be executed.

    :func:`~partial` function does a similar thing,
    but it does partial application exactly once.
    ``curry`` is a bit smarter and will do partial
    application until enough arguments passed.

    If wrong arguments are passed, ``TypeError`` will be raised immediately.

    We use a custom ``mypy`` plugin to make sure types are correct.
    Otherwise, it is currently impossible to properly type this function.

    .. code:: pycon

      >>> from returns.curry import curry

      >>> @curry
      ... def divide(number: int, by: int) -> float:
      ...     return number / by

      >>> divide(1)  # doesn't call the func and remembers arguments
      <function divide at ...>
      >>> assert divide(1)(by=10) == 0.1  # calls the func when possible
      >>> assert divide(1)(10) == 0.1  # calls the func when possible
      >>> assert divide(1, by=10) == 0.1  # or call the func like always

    Here are several examples with wrong arguments:

    .. code:: pycon

      >>> divide(1, 2, 3)
      Traceback (most recent call last):
        ...
      TypeError: too many positional arguments

      >>> divide(a=1)
      Traceback (most recent call last):
        ...
      TypeError: got an unexpected keyword argument 'a'

    Limitations:

    - It is kinda slow. Like 100 times slower than a regular function call.
    - It does not work with several builtins like ``str``, ``int``,
      and possibly other ``C`` defined callables
    - ``*args`` and ``**kwargs`` are not supported
      and we use ``Any`` as a fallback
    - Support of arguments with default values is very limited,
      because we cannot be totally sure which case we are using:
      with the default value or without it, be careful
    - We use a custom ``mypy`` plugin to make types correct,
      otherwise, it is currently impossible
    - It might not work as expected with curried ``Klass().method``,
      it might generate invalid method signature
      (looks like a bug in ``mypy``)
    - It is probably a bad idea to ``curry`` a function with lots of arguments,
      because you will end up with lots of overload functions,
      that you won't be able to understand.
      It might also be slow during the typecheck
    - Currying of ``__init__`` does not work because of the bug in ``mypy``:
      https://github.com/python/mypy/issues/8801

    We expect people to use this tool responsibly
    when they know that they are doing.

    See also:
    - https://en.wikipedia.org/wiki/Currying
    - https://stackoverflow.com/questions/218025/

    """
    argspec = Signature.from_callable(function).bind_partial()

    def decorator(*args, **kwargs):
        return _eager_curry(function, argspec, args, kwargs)

    return wraps(function)(decorator)


def _eager_curry(
    function: Callable[..., _ReturnType],
    argspec,
    args: tuple,
    kwargs: dict,
) -> _ReturnType | Callable[..., _ReturnType]:
    """
    Internal ``curry`` implementation.

    The interesting part about it is that it return the result
    or a new callable that will return a result at some point.
    """
    intermediate, full_args = _intermediate_argspec(argspec, args, kwargs)
    if full_args is not None:
        return function(*full_args[0], **full_args[1])

    # We use closures to avoid names conflict between
    # the function args and args of the curry implementation.
    def decorator(*inner_args, **inner_kwargs):
        return _eager_curry(function, intermediate, inner_args, inner_kwargs)

    return wraps(function)(decorator)


_ArgSpec: TypeAlias = (
    # Case when all arguments are bound and function can be called:
    tuple[None, tuple[tuple, dict]]
    |
    # Case when there are still unbound arguments:
    tuple[BoundArguments, None]
)


def _intermediate_argspec(
    argspec: BoundArguments,
    args: tuple,
    kwargs: dict,
) -> _ArgSpec:
    """
    That's where ``curry`` magic happens.

    We use ``Signature`` objects from ``inspect`` to bind existing arguments.

    If there's a ``TypeError`` while we ``bind`` the arguments we try again.
    The second time we try to ``bind_partial`` arguments. It can fail too!
    It fails when there are invalid arguments
    or more arguments than we can fit in a function.

    This function is slow. Any optimization ideas are welcome!
    """
    full_args = argspec.args + args
    full_kwargs = {**argspec.kwargs, **kwargs}

    try:
        argspec.signature.bind(*full_args, **full_kwargs)
    except TypeError:
        # Another option is to copy-paste and patch `getcallargs` func
        # but in this case we get responsibility to maintain it over
        # python releases.
        # This place is also responsible for raising ``TypeError`` for cases:
        # 1. When incorrect argument is provided
        # 2. When too many arguments are provided
        return argspec.signature.bind_partial(*full_args, **full_kwargs), None
    return None, (full_args, full_kwargs)



from abc import abstractmethod
from collections.abc import Callable, Sequence
from typing import ClassVar, Generic, Never, TypeVar, final

from returns.functions import compose, identity
from returns.primitives.asserts import assert_equal
from returns.primitives.hkt import KindN
from returns.primitives.laws import (
    Law,
    Law1,
    Law3,
    Lawful,
    LawSpecDef,
    law_definition,
)

_FirstType = TypeVar('_FirstType')
_SecondType = TypeVar('_SecondType')
_ThirdType = TypeVar('_ThirdType')
_UpdatedType = TypeVar('_UpdatedType')

_MappableType = TypeVar('_MappableType', bound='MappableN')

# Used in laws:
_NewType1 = TypeVar('_NewType1')
_NewType2 = TypeVar('_NewType2')


@final
class _LawSpec(LawSpecDef):
    """
    Mappable or functor laws.

    https://en.wikibooks.org/wiki/Haskell/The_Functor_class#The_functor_laws
    """

    __slots__ = ()

    @law_definition
    def identity_law(
        mappable: 'MappableN[_FirstType, _SecondType, _ThirdType]',
    ) -> None:
        """Mapping identity over a value must return the value unchanged."""
        assert_equal(mappable.map(identity), mappable)

    @law_definition
    def associative_law(
        mappable: 'MappableN[_FirstType, _SecondType, _ThirdType]',
        first: Callable[[_FirstType], _NewType1],
        second: Callable[[_NewType1], _NewType2],
    ) -> None:
        """Mapping twice or mapping a composition is the same thing."""
        assert_equal(
            mappable.map(first).map(second),
            mappable.map(compose(first, second)),
        )


class MappableN(
    Lawful['MappableN[_FirstType, _SecondType, _ThirdType]'],
    Generic[_FirstType, _SecondType, _ThirdType],
):
    """
    Allows to chain wrapped values in containers with regular functions.

    Behaves like a functor.

    See also:
        - https://en.wikipedia.org/wiki/Functor

    """

    __slots__ = ()

    _laws: ClassVar[Sequence[Law]] = (
        Law1(_LawSpec.identity_law),
        Law3(_LawSpec.associative_law),
    )

    @abstractmethod  # noqa: WPS125
    def map(
        self: _MappableType,
        function: Callable[[_FirstType], _UpdatedType],
    ) -> KindN[_MappableType, _UpdatedType, _SecondType, _ThirdType]:
        """Allows to run a pure function over a container."""


#: Type alias for kinds with one type argument.
Mappable1 = MappableN[_FirstType, Never, Never]

#: Type alias for kinds with two type arguments.
Mappable2 = MappableN[_FirstType, _SecondType, Never]

#: Type alias for kinds with three type arguments.
Mappable3 = MappableN[_FirstType, _SecondType, _ThirdType]



from collections.abc import Callable
from typing import TypeVar

from returns.interfaces.bindable import BindableN
from returns.primitives.hkt import Kinded, KindN, kinded

_FirstType = TypeVar('_FirstType')
_SecondType = TypeVar('_SecondType')
_ThirdType = TypeVar('_ThirdType')
_UpdatedType = TypeVar('_UpdatedType')

_BindableKind = TypeVar('_BindableKind', bound=BindableN)


def bind(
    function: Callable[
        [_FirstType],
        KindN[_BindableKind, _UpdatedType, _SecondType, _ThirdType],
    ],
) -> Kinded[
    Callable[
        [KindN[_BindableKind, _FirstType, _SecondType, _ThirdType]],
        KindN[_BindableKind, _UpdatedType, _SecondType, _ThirdType],
    ]
]:
    """
    Turns function's input parameter from a regular value to a container.

    In other words, it modifies the function
    signature from:
    ``a -> Container[b]``
    to:
    ``Container[a] -> Container[b]``

    Similar to :func:`returns.pointfree.lash`,
    but works for successful containers.
    This is how it should be used:

    .. code:: python

      >>> from returns.pointfree import bind
      >>> from returns.maybe import Maybe, Some, Nothing

      >>> def example(argument: int) -> Maybe[int]:
      ...     return Some(argument + 1)

      >>> assert bind(example)(Some(1)) == Some(2)
      >>> assert bind(example)(Nothing) == Nothing

    Note, that this function works for all containers with ``.bind`` method.
    See :class:`returns.primitives.interfaces.bindable.BindableN` for more info.

    """

    @kinded
    def factory(
        container: KindN[_BindableKind, _FirstType, _SecondType, _ThirdType],
    ) -> KindN[_BindableKind, _UpdatedType, _SecondType, _ThirdType]:
        return container.bind(function)

    return factory



from collections.abc import Callable
from typing import TypeVar

from returns.interfaces.specific.maybe import MaybeLikeN
from returns.primitives.hkt import Kinded, KindN, kinded

_FirstType = TypeVar('_FirstType')
_SecondType = TypeVar('_SecondType')
_ThirdType = TypeVar('_ThirdType')
_UpdatedType = TypeVar('_UpdatedType')

_MaybeLikeKind = TypeVar('_MaybeLikeKind', bound=MaybeLikeN)


def bind_optional(
    function: Callable[[_FirstType], _UpdatedType | None],
) -> Kinded[
    Callable[
        [KindN[_MaybeLikeKind, _FirstType, _SecondType, _ThirdType]],
        KindN[_MaybeLikeKind, _UpdatedType, _SecondType, _ThirdType],
    ]
]:
    """
    Binds a function returning optional value over a container.

    In other words, it modifies the function's
    signature from:
    ``a -> Optional[b]``
    to:
    ``Container[a] -> Container[b]``

    .. code:: python

      >>> from typing import Optional
      >>> from returns.pointfree import bind_optional
      >>> from returns.maybe import Some, Nothing

      >>> def example(argument: int) -> Optional[int]:
      ...     return argument + 1 if argument > 0 else None

      >>> assert bind_optional(example)(Some(1)) == Some(2)
      >>> assert bind_optional(example)(Some(0)) == Nothing
      >>> assert bind_optional(example)(Nothing) == Nothing

    Note, that this function works
    for all containers with ``.bind_optional`` method.
    See :class:`returns.primitives.interfaces.specific.maybe._MaybeLikeKind`
    for more info.

    """

    @kinded
    def factory(
        container: KindN[_MaybeLikeKind, _FirstType, _SecondType, _ThirdType],
    ) -> KindN[_MaybeLikeKind, _UpdatedType, _SecondType, _ThirdType]:
        return container.bind_optional(function)

    return factory



from collections.abc import Awaitable, Callable
from typing import TypeVar

from returns.future import FutureResult
from returns.interfaces.specific.future_result import FutureResultLikeN
from returns.primitives.hkt import Kinded, KindN, kinded

_FirstType = TypeVar('_FirstType')
_SecondType = TypeVar('_SecondType')
_ThirdType = TypeVar('_ThirdType')
_UpdatedType = TypeVar('_UpdatedType')

_FutureResultKind = TypeVar('_FutureResultKind', bound=FutureResultLikeN)


def bind_async_future_result(
    function: Callable[
        [_FirstType],
        Awaitable[FutureResult[_UpdatedType, _SecondType]],
    ],
) -> Kinded[
    Callable[
        [KindN[_FutureResultKind, _FirstType, _SecondType, _ThirdType]],
        KindN[_FutureResultKind, _UpdatedType, _SecondType, _ThirdType],
    ]
]:
    """
    Compose a container and async function returning ``FutureResult``.

    In other words, it modifies the function
    signature from:
    ``a -> Awaitable[FutureResult[b, c]]``
    to:
    ``Container[a, c] -> Container[b, c]``

    This is how it should be used:

    .. code:: python

      >>> import anyio
      >>> from returns.pointfree import bind_async_future_result
      >>> from returns.future import FutureResult
      >>> from returns.io import IOSuccess, IOFailure

      >>> async def example(argument: int) -> FutureResult[int, str]:
      ...     return FutureResult.from_value(argument + 1)

      >>> assert anyio.run(
      ...     bind_async_future_result(example)(
      ...         FutureResult.from_value(1),
      ...     ).awaitable,
      ... ) == IOSuccess(2)

      >>> assert anyio.run(
      ...     bind_async_future_result(example)(
      ...         FutureResult.from_failure('a'),
      ...     ).awaitable,
      ... ) == IOFailure('a')

    .. currentmodule: returns.primitives.interfaces.specific.future_result

    Note, that this function works
    for all containers with ``.bind_async_future`` method.
    See :class:`~FutureResultLikeN` for more info.

    """

    @kinded
    def factory(
        container: KindN[
            _FutureResultKind,
            _FirstType,
            _SecondType,
            _ThirdType,
        ],
    ) -> KindN[_FutureResultKind, _UpdatedType, _SecondType, _ThirdType]:
        return container.bind_async_future_result(function)

    return factory



from collections.abc import Iterable, Mapping
from typing import TypeAlias, cast, final

from mypy.argmap import map_actuals_to_formals
from mypy.constraints import infer_constraints_for_callable
from mypy.expandtype import expand_type
from mypy.nodes import ARG_POS, ArgKind
from mypy.plugin import FunctionContext
from mypy.types import (
    CallableType,
    FunctionLike,
    ProperType,
    TypeVarId,
    get_proper_type,
)
from mypy.types import Type as MypyType

from returns.contrib.mypy._structures.args import FuncArg
from returns.contrib.mypy._structures.types import CallableContext
from returns.contrib.mypy._typeops.analtype import analyze_call

#: Mapping of `typevar` to real type.
_Constraints: TypeAlias = Mapping[TypeVarId, MypyType]


@final
class CallableInference:
    """
    Used to infer function arguments and return type.

    There are multiple ways to do it.
    For example, one can infer argument types from its usage.
    """

    def __init__(
        self,
        case_function: CallableType,
        ctx: FunctionContext,
        *,
        fallback: CallableType | None = None,
    ) -> None:
        """
        Create the callable inference.

        Sometimes we need two functions.
        When construction one function from another
        there might be some lost information during the process.
        That's why we optionally need ``fallback``.
        If it is not provided, we treat ``case_function`` as a full one.

        Args:
            case_function: function with solved constraints.
            fallback: Function with unsolved constraints.
            ctx: Function context with checker and expr_checker objects.

        """
        self._case_function = case_function
        self._fallback = fallback or self._case_function
        self._ctx = ctx

    def from_usage(
        self,
        applied_args: list[FuncArg],
    ) -> CallableType:
        """Infers function constrains from its usage: passed arguments."""
        constraints = self._infer_constraints(applied_args)
        return expand_type(self._case_function, constraints)

    def _infer_constraints(
        self,
        applied_args: list[FuncArg],
    ) -> _Constraints:
        """Creates mapping of ``typevar`` to real type that we already know."""
        checker = self._ctx.api.expr_checker  # type: ignore
        kinds = [arg.kind for arg in applied_args]
        exprs = [arg.expression(self._ctx.context) for arg in applied_args]

        formal_to_actual = map_actuals_to_formals(
            kinds,
            [arg.name for arg in applied_args],
            self._fallback.arg_kinds,
            self._fallback.arg_names,
            lambda index: checker.accept(exprs[index]),
        )
        constraints = infer_constraints_for_callable(
            self._fallback,
            arg_types=[arg.type for arg in applied_args],
            arg_kinds=kinds,
            arg_names=[arg.name for arg in applied_args],
            formal_to_actual=formal_to_actual,
            context=checker.argument_infer_context(),
        )
        return {
            constraint.type_var: constraint.target for constraint in constraints
        }


@final
class PipelineInference:
    """
    Very helpful tool to work with functions like ``flow`` and ``pipe``.

    It iterates all over the given list of pipeline steps,
    passes the first argument, and then infers types step by step.
    """

    def __init__(self, instance: ProperType) -> None:
        """We do need the first argument to start the inference."""
        self._instance = instance

    def from_callable_sequence(
        self,
        pipeline_types: Iterable[ProperType],
        pipeline_kinds: Iterable[ArgKind],
        ctx: CallableContext,
    ) -> ProperType:
        """Pass pipeline functions to infer them one by one."""
        parameter = FuncArg(None, self._instance, ARG_POS)
        ret_type = get_proper_type(ctx.default_return_type)

        for pipeline, kind in zip(pipeline_types, pipeline_kinds, strict=False):
            ret_type = self._proper_type(
                analyze_call(
                    cast(FunctionLike, pipeline),
                    [parameter],
                    ctx,
                    show_errors=True,
                ),
            )
            parameter = FuncArg(None, ret_type, kind)
        return ret_type

    def _proper_type(self, typ: MypyType) -> ProperType:
        res_typ = get_proper_type(typ)
        if isinstance(res_typ, CallableType):
            return get_proper_type(res_typ.ret_type)
        return res_typ  # It might be `Instance` or `AnyType` or `Nothing`



from abc import abstractmethod
from collections.abc import Callable, Sequence
from typing import ClassVar, Never, TypeVar, final

from returns.interfaces import container as _container
from returns.interfaces import lashable, swappable
from returns.primitives.asserts import assert_equal
from returns.primitives.hkt import KindN
from returns.primitives.laws import (
    Law,
    Law2,
    Law3,
    Lawful,
    LawSpecDef,
    law_definition,
)

_FirstType = TypeVar('_FirstType')
_SecondType = TypeVar('_SecondType')
_ThirdType = TypeVar('_ThirdType')
_UpdatedType = TypeVar('_UpdatedType')

_SingleFailableType = TypeVar('_SingleFailableType', bound='SingleFailableN')
_DiverseFailableType = TypeVar('_DiverseFailableType', bound='DiverseFailableN')

# Used in laws:
_NewFirstType = TypeVar('_NewFirstType')


@final
class _FailableLawSpec(LawSpecDef):
    """
    Failable laws.

    We need to be sure that ``.lash`` won't lash success types.
    """

    __slots__ = ()

    @law_definition
    def lash_short_circuit_law(
        raw_value: _FirstType,
        container: 'FailableN[_FirstType, _SecondType, _ThirdType]',
        function: Callable[
            [_SecondType],
            KindN['FailableN', _FirstType, _NewFirstType, _ThirdType],
        ],
    ) -> None:
        """Ensures that you cannot lash a success."""
        assert_equal(
            container.from_value(raw_value),
            container.from_value(raw_value).lash(function),
        )


class FailableN(
    _container.ContainerN[_FirstType, _SecondType, _ThirdType],
    lashable.LashableN[_FirstType, _SecondType, _ThirdType],
    Lawful['FailableN[_FirstType, _SecondType, _ThirdType]'],
):
    """
    Base type for types that can fail.

    It is a raw type and should not be used directly.
    Use ``SingleFailableN`` and ``DiverseFailableN`` instead.
    """

    __slots__ = ()

    _laws: ClassVar[Sequence[Law]] = (
        Law3(_FailableLawSpec.lash_short_circuit_law),
    )


#: Type alias for kinds with two type arguments.
Failable2 = FailableN[_FirstType, _SecondType, Never]

#: Type alias for kinds with three type arguments.
Failable3 = FailableN[_FirstType, _SecondType, _ThirdType]


@final
class _SingleFailableLawSpec(LawSpecDef):
    """
    Single Failable laws.

    We need to be sure that ``.map`` and ``.bind``
    works correctly for ``empty`` property.
    """

    __slots__ = ()

    @law_definition
    def map_short_circuit_law(
        container: 'SingleFailableN[_FirstType, _SecondType, _ThirdType]',
        function: Callable[[_FirstType], _NewFirstType],
    ) -> None:
        """Ensures that you cannot map from the `empty` property."""
        assert_equal(
            container.empty,
            container.empty.map(function),
        )

    @law_definition
    def bind_short_circuit_law(
        container: 'SingleFailableN[_FirstType, _SecondType, _ThirdType]',
        function: Callable[
            [_FirstType],
            KindN['SingleFailableN', _NewFirstType, _SecondType, _ThirdType],
        ],
    ) -> None:
        """Ensures that you cannot bind from the `empty` property."""
        assert_equal(
            container.empty,
            container.empty.bind(function),
        )

    @law_definition
    def apply_short_circuit_law(
        container: 'SingleFailableN[_FirstType, _SecondType, _ThirdType]',
        function: Callable[[_FirstType], _NewFirstType],
    ) -> None:
        """Ensures that you cannot apply from the `empty` property."""
        wrapped_function = container.from_value(function)
        assert_equal(
            container.empty,
            container.empty.apply(wrapped_function),
        )


class SingleFailableN(
    FailableN[_FirstType, _SecondType, _ThirdType],
):
    """
    Base type for types that have just only one failed value.

    Like ``Maybe`` types where the only failed value is ``Nothing``.
    """

    __slots__ = ()

    _laws: ClassVar[Sequence[Law]] = (
        Law2(_SingleFailableLawSpec.map_short_circuit_law),
        Law2(_SingleFailableLawSpec.bind_short_circuit_law),
        Law2(_SingleFailableLawSpec.apply_short_circuit_law),
    )

    @property
    @abstractmethod
    def empty(
        self: _SingleFailableType,
    ) -> 'SingleFailableN[_FirstType, _SecondType, _ThirdType]':
        """This property represents the failed value."""


#: Type alias for kinds with two types arguments.
SingleFailable2 = SingleFailableN[_FirstType, _SecondType, Never]

#: Type alias for kinds with three type arguments.
SingleFailable3 = SingleFailableN[_FirstType, _SecondType, _ThirdType]


@final
class _DiverseFailableLawSpec(LawSpecDef):
    """
    Diverse Failable laws.

    We need to be sure that ``.map``, ``.bind``, ``.apply`` and ``.alt``
    works correctly for both success and failure types.
    """

    __slots__ = ()

    @law_definition
    def map_short_circuit_law(
        raw_value: _SecondType,
        container: 'DiverseFailableN[_FirstType, _SecondType, _ThirdType]',
        function: Callable[[_FirstType], _NewFirstType],
    ) -> None:
        """Ensures that you cannot map a failure."""
        assert_equal(
            container.from_failure(raw_value),
            container.from_failure(raw_value).map(function),
        )

    @law_definition
    def bind_short_circuit_law(
        raw_value: _SecondType,
        container: 'DiverseFailableN[_FirstType, _SecondType, _ThirdType]',
        function: Callable[
            [_FirstType],
            KindN['DiverseFailableN', _NewFirstType, _SecondType, _ThirdType],
        ],
    ) -> None:
        """
        Ensures that you cannot bind a failure.

        See: https://wiki.haskell.org/Typeclassopedia#MonadFail
        """
        assert_equal(
            container.from_failure(raw_value),
            container.from_failure(raw_value).bind(function),
        )

    @law_definition
    def apply_short_circuit_law(
        raw_value: _SecondType,
        container: 'DiverseFailableN[_FirstType, _SecondType, _ThirdType]',
        function: Callable[[_FirstType], _NewFirstType],
    ) -> None:
        """Ensures that you cannot apply a failure."""
        wrapped_function = container.from_value(function)
        assert_equal(
            container.from_failure(raw_value),
            container.from_failure(raw_value).apply(wrapped_function),
        )

    @law_definition
    def alt_short_circuit_law(
        raw_value: _SecondType,
        container: 'DiverseFailableN[_FirstType, _SecondType, _ThirdType]',
        function: Callable[[_SecondType], _NewFirstType],
    ) -> None:
        """Ensures that you cannot alt a success."""
        assert_equal(
            container.from_value(raw_value),
            container.from_value(raw_value).alt(function),
        )


class DiverseFailableN(
    FailableN[_FirstType, _SecondType, _ThirdType],
    swappable.SwappableN[_FirstType, _SecondType, _ThirdType],
    Lawful['DiverseFailableN[_FirstType, _SecondType, _ThirdType]'],
):
    """
    Base type for types that have any failed value.

    Like ``Result`` types.
    """

    __slots__ = ()

    _laws: ClassVar[Sequence[Law]] = (
        Law3(_DiverseFailableLawSpec.map_short_circuit_law),
        Law3(_DiverseFailableLawSpec.bind_short_circuit_law),
        Law3(_DiverseFailableLawSpec.apply_short_circuit_law),
        Law3(_DiverseFailableLawSpec.alt_short_circuit_law),
    )

    @classmethod
    @abstractmethod
    def from_failure(
        cls: type[_DiverseFailableType],
        inner_value: _UpdatedType,
    ) -> KindN[_DiverseFailableType, _FirstType, _UpdatedType, _ThirdType]:
        """Unit method to create new containers from any raw value."""


#: Type alias for kinds with two type arguments.
DiverseFailable2 = DiverseFailableN[_FirstType, _SecondType, Never]

#: Type alias for kinds with three type arguments.
DiverseFailable3 = DiverseFailableN[_FirstType, _SecondType, _ThirdType]



from collections.abc import Iterator
from itertools import groupby, product
from operator import itemgetter
from typing import cast, final

from mypy.nodes import ARG_STAR, ARG_STAR2
from mypy.plugin import FunctionContext
from mypy.types import (
    AnyType,
    CallableType,
    FunctionLike,
    Overloaded,
    TypeOfAny,
    get_proper_type,
)
from mypy.types import Type as MypyType

from returns.contrib.mypy._structures.args import FuncArg
from returns.contrib.mypy._typeops.transform_callable import (
    Intermediate,
    proper_type,
)

#: Raw material to build `_ArgTree`.
_RawArgTree = list[list[list[FuncArg]]]


def analyze(ctx: FunctionContext) -> MypyType:
    """Returns proper type for curried functions."""
    default_return = get_proper_type(ctx.default_return_type)
    arg_type = get_proper_type(ctx.arg_types[0][0])
    if not isinstance(arg_type, CallableType):
        return default_return
    if not isinstance(default_return, CallableType):
        return default_return

    return _CurryFunctionOverloads(arg_type, ctx).build_overloads()


@final
class _ArgTree:
    """Represents a node in tree of arguments."""

    def __init__(self, case: CallableType | None) -> None:
        self.case = case
        self.children: list[_ArgTree] = []


@final
class _CurryFunctionOverloads:
    """
    Implementation of ``@curry`` decorator typings.

    Basically does just two things:

    1. Creates all possible ordered combitations of arguments
    2. Creates ``Overload`` instances for functions' return types

    """

    def __init__(self, original: CallableType, ctx: FunctionContext) -> None:
        """
        Saving the things we need.

        Args:
            original: original function that was passed to ``@curry``.
            ctx: function context.

        """
        self._original = original
        self._ctx = ctx
        self._overloads: list[CallableType] = []
        self._args = FuncArg.from_callable(self._original)

        # We need to get rid of generics here.
        # Because, otherwise `detach_callable` with add
        # unused variables to intermediate callables.
        self._default = cast(
            CallableType,
            self._ctx.default_return_type,
        ).copy_modified(
            ret_type=AnyType(TypeOfAny.implementation_artifact),
        )

    def build_overloads(self) -> MypyType:
        """
        Builds lots of possible overloads for a given function.

        Inside we try to repsent all functions as sequence of arguments,
        grouped by the similar ones and returning one more overload instance.
        """
        if not self._args:  # There's nothing to do, function has 0 args.
            return self._original

        if any(arg.kind in {ARG_STAR, ARG_STAR2} for arg in self._args):
            # We don't support `*args` and `**kwargs`.
            # Because it is very complex. It might be fixes in the future.
            return self._default.ret_type  # Any

        argtree = self._build_argtree(
            _ArgTree(None),  # starting from root node
            list(self._slices(self._args)),
        )
        self._build_overloads_from_argtree(argtree)
        return proper_type(self._overloads)

    def _build_argtree(
        self,
        node: _ArgTree,
        source: _RawArgTree,
    ) -> '_ArgTree':
        """
        Builds argument tree.

        Each argument can point to zero, one, or more other nodes.
        Arguments that have zero children are treated as bottom (last) ones.
        Arguments that have just one child are meant to be regular functions.
        Arguments that have more than one child are treated as overloads.

        """

        def factory(
            args: _RawArgTree,
        ) -> Iterator[tuple[list[FuncArg], _RawArgTree]]:
            if not args or not args[0]:
                return  # we have reached an end of arguments
            yield from (
                (case, [group[1:] for group in grouped])
                for case, grouped in groupby(args, itemgetter(0))
            )

        for case, rest in factory(source):
            new_node = _ArgTree(
                Intermediate(self._default).with_signature(case),
            )
            node.children.append(new_node)
            self._build_argtree(source=rest, node=new_node)
        return node

    def _build_overloads_from_argtree(self, argtree: _ArgTree) -> None:
        """Generates functions from argument tree."""
        for child in argtree.children:
            self._build_overloads_from_argtree(child)
            assert child.case  # mypy is not happy

            if not child.children:
                child.case = Intermediate(child.case).with_ret_type(
                    self._original.ret_type,
                )

            if argtree.case is None:
                # Root is reached, we need to save the result:
                self._overloads.append(child.case)
            else:
                # We need to go backwards and to replace the return types
                # of the previous functions. Like so:
                # 1.  `def x -> A`
                # 2.  `def y -> A`
                # Will take `2` and apply its type to the previous function `1`.
                # Will result in `def x -> y -> A`
                # We also overloadify existing return types.
                ret_type = get_proper_type(argtree.case.ret_type)
                temp_any = (
                    isinstance(
                        ret_type,
                        AnyType,
                    )
                    and ret_type.type_of_any
                    == TypeOfAny.implementation_artifact
                )
                argtree.case = Intermediate(argtree.case).with_ret_type(
                    child.case
                    if temp_any
                    else Overloaded(
                        [child.case, *cast(FunctionLike, ret_type).items],
                    ),
                )

    def _slices(self, source: list[FuncArg]) -> Iterator[list[list[FuncArg]]]:
        """
        Generate all possible slices of a source list.

        Example::

          _slices("AB") ->
            "AB"
            "A" "B"

          _slices("ABC") ->
            "ABC"
            "AB" "C"
            "A" "BC"
            "A" "B" "C"

        """
        for doslice in product([True, False], repeat=len(source) - 1):
            slices = []
            start = 0
            for index, slicehere in enumerate(doslice, 1):
                if slicehere:
                    slices.append(source[start:index])
                    start = index
            slices.append(source[start:])
            yield slices



from collections.abc import Callable
from typing import TypeVar

from returns.contrib.hypothesis.laws import check_all_laws
from returns.interfaces import equable, mappable
from returns.primitives.container import BaseContainer, container_equality
from returns.primitives.hkt import SupportsKind1

_ValueType = TypeVar('_ValueType')
_NewValueType = TypeVar('_NewValueType')


class _Wrapper(
    BaseContainer,
    SupportsKind1['_Wrapper', _ValueType],
    mappable.Mappable1[_ValueType],
    equable.Equable,
):
    _inner_value: _ValueType

    def __init__(self, inner_value: _ValueType) -> None:
        super().__init__(inner_value)

    equals = container_equality

    def map(
        self,
        function: Callable[[_ValueType], _NewValueType],
    ) -> '_Wrapper[_NewValueType]':
        return _Wrapper(function(self._inner_value))


check_all_laws(_Wrapper, use_init=True)



from abc import abstractmethod
from collections.abc import Callable, Sequence
from typing import ClassVar, Generic, Never, TypeAlias, TypeVar, final

from returns.contrib.hypothesis.laws import check_all_laws
from returns.functions import compose, identity
from returns.primitives.asserts import assert_equal
from returns.primitives.container import BaseContainer
from returns.primitives.hkt import KindN, SupportsKind1
from returns.primitives.laws import (
    Law,
    Law1,
    Law3,
    Lawful,
    LawSpecDef,
    law_definition,
)

_FirstType = TypeVar('_FirstType')
_SecondType = TypeVar('_SecondType')
_ThirdType = TypeVar('_ThirdType')
_UpdatedType = TypeVar('_UpdatedType')

_MappableType = TypeVar('_MappableType', bound='_MappableN')
_ValueType = TypeVar('_ValueType')
_NewValueType = TypeVar('_NewValueType')

# Used in laws:
_NewType1 = TypeVar('_NewType1')
_NewType2 = TypeVar('_NewType2')


@final
class _LawSpec(LawSpecDef):
    """Copy of the functor laws for `MappableN`."""

    __slots__ = ()

    @law_definition
    def identity_law(
        mappable: '_MappableN[_FirstType, _SecondType, _ThirdType]',
    ) -> None:
        """Mapping identity over a value must return the value unchanged."""
        assert_equal(mappable.map(identity), mappable)

    @law_definition
    def associative_law(
        mappable: '_MappableN[_FirstType, _SecondType, _ThirdType]',
        first: Callable[[_FirstType], _NewType1],
        second: Callable[[_NewType1], _NewType2],
    ) -> None:
        """Mapping twice or mapping a composition is the same thing."""
        assert_equal(
            mappable.map(first).map(second),
            mappable.map(compose(first, second)),
        )


class _MappableN(
    Lawful['_MappableN[_FirstType, _SecondType, _ThirdType]'],
    Generic[_FirstType, _SecondType, _ThirdType],
):
    """Simple "user-defined" copy of `MappableN`."""

    __slots__ = ()

    _laws: ClassVar[Sequence[Law]] = (
        Law1(_LawSpec.identity_law),
        Law3(_LawSpec.associative_law),
    )

    @abstractmethod  # noqa: WPS125
    def map(
        self: _MappableType,
        function: Callable[[_FirstType], _UpdatedType],
    ) -> KindN[_MappableType, _UpdatedType, _SecondType, _ThirdType]:
        """Allows to run a pure function over a container."""


_Mappable1: TypeAlias = _MappableN[_FirstType, Never, Never]


class _ParentWrapper(_Mappable1[_ValueType]):
    """Class that is an ancestor of `_Wrapper` but has no laws."""


class _Wrapper(
    BaseContainer,
    SupportsKind1['_Wrapper', _ValueType],
    _ParentWrapper[_ValueType],
):
    """Simple instance of `_MappableN`."""

    _inner_value: _ValueType

    def __init__(self, inner_value: _ValueType) -> None:
        super().__init__(inner_value)

    def map(
        self,
        function: Callable[[_ValueType], _NewValueType],
    ) -> '_Wrapper[_NewValueType]':
        return _Wrapper(function(self._inner_value))


# We need to use `use_init=True` because `MappableN` does not automatically
# get a strategy from `strategy_from_container`.
check_all_laws(_Wrapper, use_init=True)



import pytest

from returns.future import Future, FutureResult
from returns.io import IO, IOFailure, IOSuccess
from returns.result import Failure, Success


@pytest.mark.anyio
async def test_inner_value(subtests):
    """Ensure that coroutine correct value is preserved for all units."""
    containers = [
        # We have to define these values inside the test, because
        # otherwise `anyio` will `await` reused coroutines.
        # And they have to be fresh. That's why we use subtests for it.
        FutureResult.from_value(1),
        FutureResult.from_failure(1),
        FutureResult.from_io(IO(1)),
        FutureResult.from_failed_io(IO(1)),
        FutureResult.from_ioresult(IOSuccess(1)),
        FutureResult.from_ioresult(IOFailure(1)),
        FutureResult.from_result(Success(1)),
        FutureResult.from_result(Failure(1)),
        FutureResult.from_future(Future.from_value(1)),
        FutureResult.from_failed_future(Future.from_value(1)),
        FutureResult.from_typecast(Future.from_value(Success(1))),
    ]
    for container in containers:
        with subtests.test(container=container):
            result_inst = await container
            assert result_inst._inner_value._inner_value == 1  # noqa: SLF001



import pytest

from returns.future import asyncify


@asyncify
def _function(arg: int) -> float:
    return arg / 2


@pytest.mark.anyio
async def test_asyncify_decorator():
    """Ensure that function marked with ``@asyncify`` is awaitable."""
    coro = _function(2)

    assert await coro == 1



from collections.abc import Sequence
from types import MappingProxyType
from typing import Final, Literal, overload

from mypy.checkmember import analyze_member_access
from mypy.nodes import ARG_NAMED, ARG_OPT
from mypy.types import CallableType, FunctionLike, ProperType, get_proper_type
from mypy.types import Type as MypyType

from returns.contrib.mypy._structures.args import FuncArg
from returns.contrib.mypy._structures.types import CallableContext

#: Mapping for better `call || function` argument compatibility.
_KIND_MAPPING: Final = MappingProxyType({
    # We have to replace `ARG_OPT` to `ARG_NAMED`,
    # because `ARG_OPT` is only used in function defs, not calls.
    # And `ARG_NAMED` is the same thing for calls.
    ARG_OPT: ARG_NAMED,
})


@overload
def analyze_call(
    function: FunctionLike,
    args: Sequence[FuncArg],
    ctx: CallableContext,
    *,
    show_errors: Literal[True],
) -> CallableType: ...


@overload
def analyze_call(
    function: FunctionLike,
    args: Sequence[FuncArg],
    ctx: CallableContext,
    *,
    show_errors: bool,
) -> CallableType | None: ...


def analyze_call(
    function: FunctionLike,
    args: Sequence[FuncArg],
    ctx: CallableContext,
    *,
    show_errors: bool,
) -> CallableType | None:
    """
    Analyzes function call based on passed arguments.

    Internally uses ``check_call`` from ``mypy``.
    It does a lot of magic.

    We also allow to return ``None`` instead of showing errors.
    This might be helpful for cases when we run intermediate analysis.
    """
    checker = ctx.api.expr_checker  # type: ignore[attr-defined]
    with checker.msg.filter_errors(save_filtered_errors=True) as local_errors:
        _return_type, checked_function = checker.check_call(
            function,
            [arg.expression(ctx.context) for arg in args],
            [_KIND_MAPPING.get(arg.kind, arg.kind) for arg in args],
            ctx.context,
            [arg.name for arg in args],
        )

    if not show_errors and local_errors.has_new_errors():  # noqa: WPS441
        return None

    checker.msg.add_errors(local_errors.filtered_errors())  # noqa: WPS441

    return checked_function  # type: ignore[no-any-return]


def safe_translate_to_function(
    function_def: MypyType,
    ctx: CallableContext,
) -> MypyType:
    """
    Transforms many other types to something close to callable type.

    There's why we need it:

    - We can use this on real functions
    - We can use this on ``@overload`` functions
    - We can use this on instances with ``__call__``
    - We can use this on ``Type`` types

    It can probably work with other types as well.

    This function allows us to unify this process.
    We also need to disable errors, because we explicitly pass empty args.

    This function also resolves all type arguments.
    """
    checker = ctx.api.expr_checker  # type: ignore
    with checker.msg.filter_errors():
        _return_type, function_def = checker.check_call(
            function_def,
            [],
            [],
            ctx.context,
            [],
        )
    return function_def


def translate_to_function(
    function_def: ProperType,
    ctx: CallableContext,
) -> ProperType:
    """
    Tries to translate a type into callable by accessing ``__call__`` attr.

    This might fail with ``mypy`` errors and that's how it must work.
    This also preserves all type arguments as-is.
    """
    checker = ctx.api.expr_checker  # type: ignore

    return get_proper_type(
        analyze_member_access(
            '__call__',
            function_def,
            ctx.context,
            is_lvalue=False,
            is_super=False,
            is_operator=True,
            original_type=function_def,
            chk=checker.chk,
        )
    )



from __future__ import annotations

from abc import abstractmethod
from collections.abc import Awaitable, Callable, Sequence
from typing import TYPE_CHECKING, ClassVar, TypeVar, final

from returns.interfaces.specific import future_result, reader, reader_ioresult
from returns.primitives.asserts import assert_equal
from returns.primitives.hkt import KindN
from returns.primitives.laws import (
    Law,
    Law2,
    Lawful,
    LawSpecDef,
    law_definition,
)

if TYPE_CHECKING:
    from returns.context import ReaderFutureResult  # noqa: WPS433
    from returns.future import FutureResult  # noqa: F401, WPS433

_FirstType = TypeVar('_FirstType')
_SecondType = TypeVar('_SecondType')
_ThirdType = TypeVar('_ThirdType')
_UpdatedType = TypeVar('_UpdatedType')

_ValueType = TypeVar('_ValueType')
_ErrorType = TypeVar('_ErrorType')
_EnvType = TypeVar('_EnvType')

_ReaderFutureResultLikeType = TypeVar(
    '_ReaderFutureResultLikeType',
    bound='ReaderFutureResultLikeN',
)


class ReaderFutureResultLikeN(
    reader_ioresult.ReaderIOResultLikeN[_FirstType, _SecondType, _ThirdType],
    future_result.FutureResultLikeN[_FirstType, _SecondType, _ThirdType],
):
    """
    Interface for all types that do look like ``ReaderFutureResult`` instance.

    Cannot be called.
    """

    __slots__ = ()

    @abstractmethod
    def bind_context_future_result(
        self: _ReaderFutureResultLikeType,
        function: Callable[
            [_FirstType],
            ReaderFutureResult[_UpdatedType, _SecondType, _ThirdType],
        ],
    ) -> KindN[
        _ReaderFutureResultLikeType,
        _UpdatedType,
        _SecondType,
        _ThirdType,
    ]:
        """Bind a ``ReaderFutureResult`` returning function over a container."""

    @abstractmethod
    def bind_async_context_future_result(
        self: _ReaderFutureResultLikeType,
        function: Callable[
            [_FirstType],
            Awaitable[
                ReaderFutureResult[_UpdatedType, _SecondType, _ThirdType],
            ],
        ],
    ) -> KindN[
        _ReaderFutureResultLikeType,
        _UpdatedType,
        _SecondType,
        _ThirdType,
    ]:
        """Bind async ``ReaderFutureResult`` function."""

    @classmethod
    @abstractmethod
    def from_future_result_context(
        cls: type[_ReaderFutureResultLikeType],
        inner_value: ReaderFutureResult[_ValueType, _ErrorType, _EnvType],
    ) -> KindN[_ReaderFutureResultLikeType, _ValueType, _ErrorType, _EnvType]:
        """Unit method to create new containers from ``ReaderFutureResult``."""


#: Type alias for kinds with three type arguments.
ReaderFutureResultLike3 = ReaderFutureResultLikeN[
    _FirstType,
    _SecondType,
    _ThirdType,
]


@final
class _LawSpec(LawSpecDef):
    """
    Concrete laws for ``ReaderFutureResultBasedN``.

    See: https://github.com/haskell/mtl/pull/61/files
    """

    __slots__ = ()

    @law_definition
    def asking_law(
        container: ReaderFutureResultBasedN[
            _FirstType, _SecondType, _ThirdType
        ],
        env: _ThirdType,
    ) -> None:
        """Asking for an env, always returns the env."""
        assert_equal(
            container.ask().__call__(env),  # noqa: PLC2801
            container.from_value(env).__call__(env),  # noqa: PLC2801
        )


class ReaderFutureResultBasedN(
    ReaderFutureResultLikeN[_FirstType, _SecondType, _ThirdType],
    reader.CallableReader3[
        _FirstType,
        _SecondType,
        _ThirdType,
        # Calls:
        'FutureResult[_FirstType, _SecondType]',
        _ThirdType,
    ],
    Lawful['ReaderFutureResultBasedN[_FirstType, _SecondType, _ThirdType]'],
):
    """
    This interface is very specific to our ``ReaderFutureResult`` type.

    The only thing that differs from ``ReaderFutureResultLikeN``
    is that we know the specific types for its ``__call__`` method.

    In this case the return type of ``__call__`` is ``FutureResult``.
    """

    __slots__ = ()

    _laws: ClassVar[Sequence[Law]] = (Law2(_LawSpec.asking_law),)


#: Type alias for kinds with three type arguments.
ReaderFutureResultBased3 = ReaderFutureResultBasedN[
    _FirstType,
    _SecondType,
    _ThirdType,
]



from collections.abc import Callable
from typing import TypeVar, final

from returns.interfaces import bindable, equable, lashable, swappable
from returns.primitives.container import BaseContainer, container_equality
from returns.primitives.hkt import Kind2, SupportsKind2, dekind

_FirstType = TypeVar('_FirstType')
_SecondType = TypeVar('_SecondType')

_NewFirstType = TypeVar('_NewFirstType')
_NewSecondType = TypeVar('_NewSecondType')


@final
class Pair(
    BaseContainer,
    SupportsKind2['Pair', _FirstType, _SecondType],
    bindable.Bindable2[_FirstType, _SecondType],
    swappable.Swappable2[_FirstType, _SecondType],
    lashable.Lashable2[_FirstType, _SecondType],
    equable.Equable,
):
    """
    A type that represents a pair of something.

    Like to coordinates ``(x, y)`` or two best friends.
    Or a question and an answer.

    """

    def __init__(
        self,
        inner_value: tuple[_FirstType, _SecondType],
    ) -> None:
        """Saves passed tuple as ``._inner_value`` inside this instance."""
        super().__init__(inner_value)

    # `Equable` part:

    equals = container_equality  # we already have this defined for all types

    # `Mappable` part via `BiMappable`:

    def map(
        self,
        function: Callable[[_FirstType], _NewFirstType],
    ) -> 'Pair[_NewFirstType, _SecondType]':
        return Pair((function(self._inner_value[0]), self._inner_value[1]))

    # `BindableN` part:

    def bind(
        self,
        function: Callable[
            [_FirstType],
            Kind2['Pair', _NewFirstType, _SecondType],
        ],
    ) -> 'Pair[_NewFirstType, _SecondType]':
        return dekind(function(self._inner_value[0]))

    # `AltableN` part via `BiMappableN`:

    def alt(
        self,
        function: Callable[[_SecondType], _NewSecondType],
    ) -> 'Pair[_FirstType, _NewSecondType]':
        return Pair((self._inner_value[0], function(self._inner_value[1])))

    # `LashableN` part:

    def lash(
        self,
        function: Callable[
            [_SecondType],
            Kind2['Pair', _FirstType, _NewSecondType],
        ],
    ) -> 'Pair[_FirstType, _NewSecondType]':
        return dekind(function(self._inner_value[1]))

    # `SwappableN` part:

    def swap(self) -> 'Pair[_SecondType, _FirstType]':
        return Pair((self._inner_value[1], self._inner_value[0]))



from returns.result import Failure, Success, safe


@safe
def div(first_number: int, second_number: int) -> int:
    return first_number // second_number


match div(1, 0):
    # Matches if the result stored inside `Success` is `10`
    case Success(10):
        print('Result is "10"')

    # Matches any `Success` instance and binds its value to the `value` variable
    case Success(value):
        print(f'Result is "{value}"')

    # Matches if the result stored inside `Failure` is `ZeroDivisionError`
    case Failure(ZeroDivisionError()):
        print('"ZeroDivisionError" was raised')

    # Matches any `Failure` instance
    case Failure(_):
        print('The division was a failure')



from collections.abc import Iterable
from typing import TypeVar

from returns.interfaces.unwrappable import Unwrappable
from returns.primitives.exceptions import UnwrapFailedError

_ValueType_co = TypeVar('_ValueType_co', covariant=True)
_ErrorType_co = TypeVar('_ErrorType_co', covariant=True)


def partition(
    containers: Iterable[Unwrappable[_ValueType_co, _ErrorType_co],],
) -> tuple[list[_ValueType_co], list[_ErrorType_co]]:
    """
    Partition a list of unwrappables into successful and failed values.

    Preserves order.

    .. code:: python

        >>> from returns.result import Failure, Success
        >>> from returns.methods import partition

        >>> results = [Success(1), Failure(2), Success(3), Failure(4)]
        >>> partition(results)
        ([1, 3], [2, 4])

    """
    successes: list[_ValueType_co] = []
    failures: list[_ErrorType_co] = []
    for container in containers:
        try:
            successes.append(container.unwrap())
        except UnwrapFailedError:
            failures.append(container.failure())
    return successes, failures



"""This module was quite a big one, so we have split it."""

from returns.context.requires_context import NoDeps as NoDeps
from returns.context.requires_context import Reader as Reader
from returns.context.requires_context import RequiresContext as RequiresContext
from returns.context.requires_context_future_result import (
    ReaderFutureResult as ReaderFutureResult,
)
from returns.context.requires_context_future_result import (
    ReaderFutureResultE as ReaderFutureResultE,
)
from returns.context.requires_context_future_result import (
    RequiresContextFutureResult as RequiresContextFutureResult,
)
from returns.context.requires_context_future_result import (
    RequiresContextFutureResultE as RequiresContextFutureResultE,
)
from returns.context.requires_context_ioresult import (
    ReaderIOResult as ReaderIOResult,
)
from returns.context.requires_context_ioresult import (
    ReaderIOResultE as ReaderIOResultE,
)
from returns.context.requires_context_ioresult import (
    RequiresContextIOResult as RequiresContextIOResult,
)
from returns.context.requires_context_ioresult import (
    RequiresContextIOResultE as RequiresContextIOResultE,
)
from returns.context.requires_context_result import ReaderResult as ReaderResult
from returns.context.requires_context_result import (
    ReaderResultE as ReaderResultE,
)
from returns.context.requires_context_result import (
    RequiresContextResult as RequiresContextResult,
)
from returns.context.requires_context_result import (
    RequiresContextResultE as RequiresContextResultE,
)



import pytest

from returns.io import IO, IOFailure, IOResult, IOSuccess
from returns.maybe import Maybe, Nothing, Some
from returns.result import Failure, Result, Success


@pytest.mark.parametrize(
    'container',
    [
        Success(10),
        Success(42),
        Failure(RuntimeError()),
        Failure(Exception()),
    ],
)
def test_result_pattern_matching(container: Result[int, Exception]):
    """Ensures ``Result`` containers work properly with pattern matching."""
    match container:
        case Success(10):
            assert isinstance(container, Success)
            assert container.unwrap() == 10
        case Success(value):
            assert isinstance(container, Success)
            assert value == 42
            assert container.unwrap() == value
        case Failure(RuntimeError()):
            assert isinstance(container, Failure)
            assert isinstance(container.failure(), RuntimeError)
        case Failure(_):
            assert isinstance(container, Failure)
            assert isinstance(container.failure(), Exception)
        case _:
            pytest.fail('Was not matched')


@pytest.mark.parametrize(
    'container',
    [
        Some('SOME'),
        Some('THERE IS SOME VALUE'),
        Nothing,
    ],
)
def test_maybe_pattern_matching(container: Maybe[str]):
    """Ensures ``Maybe`` containers work properly with pattern matching."""
    match container:
        case Some('SOME'):
            assert isinstance(container, Some)
            assert container.unwrap() == 'SOME'
        case Some(value):
            assert isinstance(container, Some)
            assert value == 'THERE IS SOME VALUE'
            assert container.unwrap() == value
        case Maybe.empty:
            assert container is Nothing
        case _:
            pytest.fail('Was not matched')


@pytest.mark.parametrize(
    'container',
    [
        IOSuccess(42.0),
        IOSuccess(10.0),
        IOFailure(50),
    ],
)
def test_ioresult_pattern_matching(container: IOResult[float, int]):
    """Ensures ``IOResult`` containers work properly with pattern matching."""
    match container:
        case IOSuccess(Success(42.0)):
            assert isinstance(container, IOSuccess)
            assert container.unwrap() == IO(42.0)
        case IOSuccess(value):
            assert isinstance(container, IOSuccess)
            assert container.unwrap() == IO(value.unwrap())
        case IOFailure(_):
            assert isinstance(container, IOFailure)
            assert container.failure() == IO(50)
        case _:
            pytest.fail('Was not matched')



from typing import Any, Never, Self

from returns.primitives.exceptions import ImmutableStateError


class Immutable:
    """
    Helper type for objects that should be immutable.

    When applied, each instance becomes immutable.
    Nothing can be added or deleted from it.

    .. code:: pycon
      :force:

      >>> from returns.primitives.types import Immutable
      >>> class MyModel(Immutable):
      ...     ...

      >>> model = MyModel()
      >>> model.prop = 1
      Traceback (most recent call last):
         ...
      returns.primitives.exceptions.ImmutableStateError

    See :class:`returns.primitives.container.BaseContainer` for examples.

    """

    __slots__ = ()

    def __copy__(self) -> Self:
        """Returns itself."""
        return self

    def __deepcopy__(self, memo: dict[Any, Any]) -> Self:
        """Returns itself."""
        return self

    def __setattr__(self, attr_name: str, attr_value: Any) -> Never:
        """Makes inner state of the containers immutable for modification."""
        raise ImmutableStateError

    def __delattr__(self, attr_name: str) -> Never:  # noqa: WPS603
        """Makes inner state of the containers immutable for deletion."""
        raise ImmutableStateError



from returns.io import IO, impure


def _fake_impure_function(some_param: int) -> int:
    return some_param


def test_impure():
    """Ensures that impure returns IO container."""
    impure_result = impure(_fake_impure_function)(1)
    assert isinstance(impure_result, IO)
    assert impure_result == IO(1)



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, Any, TypeVar

from hypothesis import strategies as st

if TYPE_CHECKING:
    from returns.primitives.laws import Lawful


def strategy_from_container(
    container_type: type[Lawful],
    *,
    use_init: bool = False,
) -> Callable[[type], st.SearchStrategy]:
    """
    Creates a strategy from a container type.

    Basically, containers should not support ``__init__``
    even when they have one.
    Because, that can be very complex: for example ``FutureResult`` requires
    ``Awaitable[Result[a, b]]`` as an ``__init__`` value.

    But, custom containers pass ``use_init``
    if they are not an instance of ``ApplicativeN``
    and do not have a working ``.from_value`` method.
    For example, pure ``MappableN`` can do that.

    We also try to resolve generic arguments.
    So, ``Result[_ValueType, Exception]``
    will produce any value for success cases
    and only exceptions for failure cases.
    """
    from returns.interfaces.applicative import ApplicativeN  # noqa: PLC0415
    from returns.interfaces.specific import maybe, result  # noqa: PLC0415

    def factory(type_: type) -> st.SearchStrategy:
        value_type, error_type = _get_type_vars(type_)

        strategies: list[st.SearchStrategy[Any]] = []
        if use_init and getattr(container_type, '__init__', None):
            strategies.append(st.builds(container_type))
        if issubclass(container_type, ApplicativeN):
            strategies.append(
                st.builds(
                    container_type.from_value,
                    st.from_type(value_type),
                )
            )
        if issubclass(container_type, result.ResultLikeN):
            strategies.append(
                st.builds(
                    container_type.from_failure,
                    st.from_type(error_type),
                )
            )
        if issubclass(container_type, maybe.MaybeLikeN):
            strategies.append(
                st.builds(
                    container_type.from_optional,
                    st.from_type(value_type),
                )
            )
        return st.one_of(*strategies)

    return factory


_FirstType = TypeVar('_FirstType')
_SecondType = TypeVar('_SecondType')


def _get_type_vars(thing: type):
    return getattr(thing, '__args__', (_FirstType, _SecondType))[:2]



from collections.abc import Callable
from typing import TypeVar

from returns.interfaces.applicative import ApplicativeN
from returns.primitives.hkt import Kinded, KindN, kinded

_FirstType = TypeVar('_FirstType')
_SecondType = TypeVar('_SecondType')
_ThirdType = TypeVar('_ThirdType')
_UpdatedType = TypeVar('_UpdatedType')

_ApplicativeKind = TypeVar('_ApplicativeKind', bound=ApplicativeN)


def apply(
    container: KindN[
        _ApplicativeKind,
        Callable[[_FirstType], _UpdatedType],
        _SecondType,
        _ThirdType,
    ],
) -> Kinded[
    Callable[
        [KindN[_ApplicativeKind, _FirstType, _SecondType, _ThirdType]],
        KindN[_ApplicativeKind, _UpdatedType, _SecondType, _ThirdType],
    ]
]:
    """
    Turns container containing a function into a callable.

    In other words, it modifies the function
    signature from:
    ``Container[a -> b]``
    to:
    ``Container[a] -> Container[b]``

    This is how it should be used:

    .. code:: python

      >>> from returns.pointfree import apply
      >>> from returns.maybe import Some, Nothing

      >>> def example(argument: int) -> int:
      ...     return argument + 1

      >>> assert apply(Some(example))(Some(1)) == Some(2)
      >>> assert apply(Some(example))(Nothing) == Nothing
      >>> assert apply(Nothing)(Some(1)) == Nothing
      >>> assert apply(Nothing)(Nothing) == Nothing

    Note, that this function works for all containers with ``.apply`` method.
    See :class:`returns.interfaces.applicative.ApplicativeN` for more info.

    """

    @kinded
    def factory(
        other: KindN[_ApplicativeKind, _FirstType, _SecondType, _ThirdType],
    ) -> KindN[_ApplicativeKind, _UpdatedType, _SecondType, _ThirdType]:
        return other.apply(container)

    return factory



from collections.abc import Sequence
from enum import StrEnum, unique

from mypy.checkmember import analyze_member_access
from mypy.plugin import (
    AttributeContext,
    FunctionContext,
    MethodContext,
    MethodSigContext,
)
from mypy.typeops import bind_self
from mypy.types import (
    AnyType,
    CallableType,
    FunctionLike,
    Instance,
    Overloaded,
    TypeOfAny,
    TypeType,
    TypeVarType,
    get_proper_type,
)
from mypy.types import Type as MypyType

from returns.contrib.mypy._typeops.fallback import asserts_fallback_to_any
from returns.contrib.mypy._typeops.visitor import translate_kind_instance

# TODO: probably we can validate `KindN[]` creation during `get_analtype`


@asserts_fallback_to_any
def attribute_access(ctx: AttributeContext) -> MypyType:
    """
    Ensures that attribute access to ``KindN`` is correct.

    In other words:

    .. code:: python

        from typing import TypeVar
        from returns.primitives.hkt import KindN
        from returns.interfaces.mappable import MappableN

        _MappableType = TypeVar('_MappableType', bound=MappableN)

        kind: KindN[_MappableType, int, int, int]
        reveal_type(kind.map)  # will work correctly!

    """
    assert isinstance(ctx.type, Instance)
    instance = get_proper_type(ctx.type.args[0])

    if isinstance(instance, TypeVarType):
        bound = get_proper_type(instance.upper_bound)
        assert isinstance(bound, Instance)
        accessed = bound.copy_modified(
            args=_crop_kind_args(ctx.type, bound.args),
        )
    elif isinstance(instance, Instance):
        accessed = instance.copy_modified(args=_crop_kind_args(ctx.type))
    else:
        return ctx.default_attr_type

    return analyze_member_access(
        ctx.context.name,  # type: ignore
        accessed,
        ctx.context,
        is_lvalue=False,
        is_super=False,
        is_operator=False,
        original_type=instance,
        chk=ctx.api,  # type: ignore
        in_literal_context=ctx.api.expr_checker.is_literal_context(),  # type: ignore
    )


def dekind(ctx: FunctionContext) -> MypyType:
    """
    Infers real type behind ``Kind`` form.

    Basically, it turns ``Kind[IO, int]`` into ``IO[int]``.
    The only limitation is that it works with
    only ``Instance`` type in the first type argument position.

    So, ``dekind(KindN[T, int])`` will fail.
    """
    kind = get_proper_type(ctx.arg_types[0][0])
    assert isinstance(kind, Instance)  # mypy requires these lines

    kind_inst = get_proper_type(kind.args[0])

    if not isinstance(kind_inst, Instance):
        ctx.api.fail(_KindErrors.dekind_not_instance, ctx.context)
        return AnyType(TypeOfAny.from_error)

    return kind_inst.copy_modified(args=_crop_kind_args(kind))


@asserts_fallback_to_any
def kinded_signature(ctx: MethodSigContext) -> CallableType:
    """
    Returns the internal function wrapped as ``Kinded[def]``.

    Works for ``Kinded`` class when ``__call__`` magic method is used.
    See :class:`returns.primitives.hkt.Kinded` for more information.
    """
    assert isinstance(ctx.type, Instance)

    wrapped_method = get_proper_type(ctx.type.args[0])
    assert isinstance(wrapped_method, FunctionLike)

    if isinstance(wrapped_method, Overloaded):
        return ctx.default_signature

    assert isinstance(wrapped_method, CallableType)
    return wrapped_method


# TODO: we should raise an error if bound type does not have any `KindN`
# instances, because that's not how `@kinded` and `Kinded[]` should be used.
def kinded_call(ctx: MethodContext) -> MypyType:
    """
    Reveals the correct return type of ``Kinded.__call__`` method.

    Turns ``-> KindN[I, t1, t2, t3]`` into ``-> I[t1, t2, t3]``.

    Also strips unused type arguments for ``KindN``, so:
    - ``KindN[IO, int, Never, Never]`` will be ``IO[int]``
    - ``KindN[Result, int, str, Never]`` will be ``Result[int, str]``

    It also processes nested ``KindN`` with recursive strategy.

    See :class:`returns.primitives.hkt.Kinded` for more information.
    """
    return translate_kind_instance(ctx.default_return_type)


@asserts_fallback_to_any
def kinded_get_descriptor(ctx: MethodContext) -> MypyType:
    """
    Used to analyze ``@kinded`` method calls.

    We do this due to ``__get__`` descriptor magic.
    """
    assert isinstance(ctx.type, Instance)

    wrapped_method = get_proper_type(ctx.type.args[0])
    assert isinstance(wrapped_method, CallableType)

    self_type = get_proper_type(wrapped_method.arg_types[0])
    signature = bind_self(
        wrapped_method,
        is_classmethod=isinstance(self_type, TypeType),
    )
    return ctx.type.copy_modified(args=[signature])


@unique
class _KindErrors(StrEnum):
    """Represents a set of possible errors we can throw during typechecking."""

    dekind_not_instance = (
        'dekind must be used with Instance as the first type argument'
    )


def _crop_kind_args(
    kind: Instance,
    limit: Sequence[MypyType] | None = None,
) -> tuple[MypyType, ...]:
    """Returns the correct amount of type arguments for a kind."""
    if limit is None:
        limit = kind.args[0].args  # type: ignore
    return kind.args[1 : len(limit) + 1]



import types
from collections.abc import Callable, Iterator
from contextlib import AbstractContextManager, contextmanager
from inspect import FrameInfo, stack
from typing import TypeVar, overload

from returns.result import Failure

_FunctionType = TypeVar('_FunctionType', bound=Callable)


@overload
def collect_traces() -> AbstractContextManager[None]: ...


@overload
def collect_traces(function: _FunctionType) -> _FunctionType: ...


def collect_traces(
    function: _FunctionType | None = None,
) -> _FunctionType | AbstractContextManager[None]:
    """
    Context Manager/Decorator to active traces collect to the Failures.

    .. code:: python

        >>> from inspect import FrameInfo

        >>> from returns.io import IOResult
        >>> from returns.result import Result
        >>> from returns.primitives.tracing import collect_traces

        >>> with collect_traces():
        ...     traced_failure = Result.from_failure('Traced Failure')
        >>> non_traced_failure = IOResult.from_failure('Non Traced Failure')

        >>> assert non_traced_failure.trace is None
        >>> assert isinstance(traced_failure.trace, list)
        >>> assert all(
        ...     isinstance(trace_line, FrameInfo)
        ...     for trace_line in traced_failure.trace
        ... )

        >>> for trace_line in traced_failure.trace:
        ...     print(  # doctest: +SKIP
        ...         '{0}:{1} in `{2}`'.format(
        ...             trace_line.filename,
        ...             trace_line.lineno,
        ...             trace_line.function,
        ...         ),
        ...     )
        ...
        /returns/returns/result.py:525 in `Failure`
        /returns/returns/result.py:322 in `from_failure`
        /example_folder/example.py:1 in `<module>`
        # doctest: # noqa: DAR301, E501

    """

    @contextmanager
    def factory() -> Iterator[None]:
        unpatched_get_trace = getattr(Failure, '_get_trace')  # noqa: B009
        substitute_get_trace = types.MethodType(_get_trace, Failure)
        setattr(Failure, '_get_trace', substitute_get_trace)  # noqa: B010
        try:  # noqa: WPS501
            yield
        finally:
            setattr(Failure, '_get_trace', unpatched_get_trace)  # noqa: B010

    return factory()(function) if function else factory()


def _get_trace(_self: Failure) -> list[FrameInfo] | None:
    """
    Function to be used on Monkey Patching.

    This function is the substitute for '_get_trace' method from ``Failure``
    class on Monkey Patching promoted by
    :func:`returns.primitives.tracing.collect_traces` function.

    We get all the call stack from the current call and return it from the
    third position, to avoid two useless calls on the call stack.
    Those useless calls are a call to this function and a call to `__init__`
    method from ``Failure`` class. We're just interested in the call stack
    ending on ``Failure`` function call!

    See also:
        - https://github.com/dry-python/returns/issues/409

    """
    current_stack = stack()
    return current_stack[2:]



from collections.abc import Callable
from typing import TypeVar

import pytest
from hypothesis.errors import ResolutionFailed

from returns.contrib.hypothesis.laws import check_all_laws
from returns.interfaces import mappable
from returns.primitives.container import BaseContainer
from returns.primitives.hkt import SupportsKind1

pytestmark = pytest.mark.xfail(raises=ResolutionFailed)

_ValueType = TypeVar('_ValueType')
_NewValueType = TypeVar('_NewValueType')


class _WithInitNoFlag(
    BaseContainer,
    SupportsKind1['_WithInitNoFlag', _ValueType],
    mappable.Mappable1[_ValueType],
):
    """Does not have any ways to be constructed."""

    def map(
        self,
        function: Callable[[_ValueType], _NewValueType],
    ) -> '_WithInitNoFlag[_NewValueType]':
        """We need `map` to have `laws`, should not be called."""
        raise NotImplementedError


check_all_laws(_WithInitNoFlag)



import anyio
import pytest

from returns.future import FutureResult, future_safe
from returns.io import IOFailure, IOSuccess


@future_safe
async def _coro(arg: int) -> float:
    assert isinstance(arg, int)
    await anyio.sleep(0)
    return 1 / arg


@future_safe(exceptions=(ZeroDivisionError,))
async def _coro_two(arg: int) -> float:
    assert isinstance(arg, int)
    await anyio.sleep(0)
    return 1 / arg


@future_safe((ZeroDivisionError,))
async def _coro_three(arg: int | str) -> float:
    assert isinstance(arg, int)
    await anyio.sleep(0)
    return 1 / arg


@pytest.mark.anyio
async def test_future_safe_decorator():
    """Ensure that coroutine marked with ``@future_safe``."""
    future_instance = _coro(2)

    assert isinstance(future_instance, FutureResult)
    assert await future_instance == IOSuccess(0.5)


@pytest.mark.anyio
async def test_future_safe_decorator_failure():
    """Ensure that coroutine marked with ``@future_safe``."""
    future_instance = _coro(0)

    assert isinstance(future_instance, FutureResult)
    assert isinstance(await future_instance, IOFailure)


@pytest.mark.anyio
async def test_future_safe_decorator_w_expected_error(subtests):
    """Ensure that coroutine marked with ``@future_safe``."""
    expected = '<IOResult: <Failure: division by zero>>'

    for future_instance in (_coro_two(0), _coro_three(0)):
        with subtests.test(future_instance=future_instance):
            assert isinstance(future_instance, FutureResult)
            inner_result = await future_instance
            assert str(inner_result) == expected


@pytest.mark.anyio
@pytest.mark.xfail(raises=AssertionError)
async def test_future_safe_decorator_w_unexpected_error():
    """Ensure that coroutine marked with ``@future_safe``."""
    await _coro_three('0')



import pickle  # noqa: S403

from returns.maybe import Nothing
from returns.primitives.exceptions import UnwrapFailedError
from returns.result import Failure


def test_pickle_unwrap_failed_error_from_maybe():
    """Ensures that UnwrapFailedError with Maybe can be pickled."""
    serialized = None
    try:
        Nothing.unwrap()  # This will raise UnwrapFailedError
    except UnwrapFailedError as error:
        serialized = pickle.dumps(error)

    deserialized_error = pickle.loads(serialized)  # noqa: S301
    assert deserialized_error.halted_container == Nothing


def test_pickle_unwrap_failed_error_from_result():
    """Ensures that UnwrapFailedError with Result can be pickled."""
    serialized = None
    try:
        Failure('error').unwrap()  # This will raise UnwrapFailedError
    except UnwrapFailedError as error:
        serialized = pickle.dumps(error)

    deserialized_error = pickle.loads(serialized)  # noqa: S301
    assert deserialized_error.halted_container == Failure('error')



from collections.abc import Callable
from typing import TypeVar

from returns.interfaces.altable import AltableN
from returns.primitives.hkt import Kinded, KindN, kinded

_FirstType = TypeVar('_FirstType')
_SecondType = TypeVar('_SecondType')
_ThirdType = TypeVar('_ThirdType')
_UpdatedType = TypeVar('_UpdatedType')

_AltableKind = TypeVar('_AltableKind', bound=AltableN)


def alt(
    function: Callable[[_SecondType], _UpdatedType],
) -> Kinded[
    Callable[
        [KindN[_AltableKind, _FirstType, _SecondType, _ThirdType]],
        KindN[_AltableKind, _FirstType, _UpdatedType, _ThirdType],
    ]
]:
    """
    Lifts function to be wrapped in a container for better composition.

    In other words, it modifies the function's
    signature from:
    ``a -> b``
    to:
    ``Container[a] -> Container[b]``

    This is how it should be used:

    .. code:: python

        >>> from returns.io import IOFailure, IOSuccess
        >>> from returns.pointfree import alt

        >>> def example(argument: int) -> float:
        ...     return argument / 2

        >>> assert alt(example)(IOSuccess(1)) == IOSuccess(1)
        >>> assert alt(example)(IOFailure(4)) == IOFailure(2.0)

    Note, that this function works for all containers with ``.alt`` method.
    See :class:`returns.primitives.interfaces.altable.AltableN` for more info.

    """

    @kinded
    def factory(
        container: KindN[_AltableKind, _FirstType, _SecondType, _ThirdType],
    ) -> KindN[_AltableKind, _FirstType, _UpdatedType, _ThirdType]:
        return container.alt(function)

    return factory



from collections.abc import Callable
from typing import TypeVar

from returns.interfaces.specific.ioresult import IOResultLikeN
from returns.primitives.hkt import Kind3, Kinded, kinded
from returns.result import Result

_FirstType = TypeVar('_FirstType')
_NewFirstType = TypeVar('_NewFirstType')
_SecondType = TypeVar('_SecondType')
_ThirdType = TypeVar('_ThirdType')

_IOResultLikeKind = TypeVar('_IOResultLikeKind', bound=IOResultLikeN)


def compose_result(
    function: Callable[
        [Result[_FirstType, _SecondType]],
        Kind3[_IOResultLikeKind, _NewFirstType, _SecondType, _ThirdType],
    ],
) -> Kinded[
    Callable[
        [Kind3[_IOResultLikeKind, _FirstType, _SecondType, _ThirdType]],
        Kind3[_IOResultLikeKind, _NewFirstType, _SecondType, _ThirdType],
    ]
]:
    """
    Composes inner ``Result`` with ``IOResultLike`` returning function.

    Can be useful when you need an access to both states of the result.

    .. code:: python

      >>> from returns.io import IOResult, IOSuccess, IOFailure
      >>> from returns.pointfree import compose_result
      >>> from returns.result import Result

      >>> def modify_string(container: Result[str, str]) -> IOResult[str, str]:
      ...     return IOResult.from_result(
      ...         container.map(str.upper).alt(str.lower),
      ...     )

      >>> assert compose_result(modify_string)(
      ...     IOSuccess('success')
      ... ) == IOSuccess('SUCCESS')
      >>> assert compose_result(modify_string)(
      ...     IOFailure('FAILURE')
      ... ) == IOFailure('failure')

    """

    @kinded
    def factory(
        container: Kind3[
            _IOResultLikeKind,
            _FirstType,
            _SecondType,
            _ThirdType,
        ],
    ) -> Kind3[_IOResultLikeKind, _NewFirstType, _SecondType, _ThirdType]:
        return container.compose_result(function)

    return factory
